(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2016, 2017)
Author:  Sarah Winkler <sarah.winkler@uibk.ac.at> (2017)
License: LGPL (see file COPYING.LESSER)
*)

section \<open>Ordered Completion\<close>

theory Ordered_Completion
  imports
    "../Rewriting/Ordered_Rewriting"
    "../Normalization_Equivalence/Encompassment"
    "../Normalization_Equivalence/Normalization_Equivalence"
    "../Orderings/Lexicographic_Extension"
    "../Orderings/Multiset_Extension"
    "../Auxiliaries/Multiset_Extension2"
    "../Abstract_Completion/Abstract_Completion"
    "../Abstract_Completion/Prime_Critical_Pairs"
begin

locale ordered_completion = reduction_order
begin

definition "ostep E R = rstep R \<union> ordstep {\<succ>} (E\<^sup>\<leftrightarrow>)"

lemma ostep_iff_ordstep:
  assumes "R \<subseteq> {\<succ>}"
  shows "ostep E R = ordstep {\<succ>} (R \<union> E\<^sup>\<leftrightarrow>)"
  using subst_closed_less and assms by (simp add: ordstep_Un ostep_def ordstep_rstep_conv)

lemma ostep_cases [consumes 1, cases pred: ostep]:
  assumes "(s, t) \<in> ostep E R"
  obtains (rstep) l r C \<sigma> where "(l, r) \<in> R" and "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  | (ordstep) l r C \<sigma> where "(l, r) \<in> E\<^sup>\<leftrightarrow>" and "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma>" and "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  using assms by (fastforce simp: ostep_def elim: ordstep.cases)

lemma ostep_mono:
  "E \<subseteq> E' \<Longrightarrow> R \<subseteq> R'\<Longrightarrow> ostep E R \<subseteq> ostep E' R'"
  by (auto simp: ostep_def rstep_mono [THEN subsetD]
      intro: ordstep_mono [THEN subsetD, OF _ subset_refl, of "E\<^sup>\<leftrightarrow>" "E'\<^sup>\<leftrightarrow>"])

lemma rstep_imp_ostep: "(s, t) \<in> rstep R \<Longrightarrow> (s, t) \<in> ostep E R" by (auto simp: ostep_def)

lemma ostep_imp_rstep_sym:
  "(s, t) \<in> ostep E R \<Longrightarrow> (s, t) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>)"
  unfolding ostep_def using ordstep_imp_rstep by fast

lemma E_imp_ostep:
  "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma> \<Longrightarrow> (l, r) \<in> E\<^sup>\<leftrightarrow> \<Longrightarrow> s = C\<langle>l \<cdot> \<sigma>\<rangle> \<Longrightarrow> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<Longrightarrow> (s, t) \<in> ostep E R"
  by (auto simp: ostep_def intro: ordstep.intros)

lemma ostep_subst:
  "(s, t) \<in> ostep E R \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ostep E R"
  using ordstep_subst [of "{\<succ>}", OF subst_closed_less] and rstep.subst
  by (auto simp: ostep_def)

lemma ostep_ctxt:
  "(s, t) \<in> ostep E R \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> ostep E R"
  using ordstep_ctxt [where ord = "{\<succ>}"] and rstep.ctxt
  by (auto simp: ostep_def)

lemma ostep_imp_less:
  "R \<subseteq> {\<succ>} \<Longrightarrow> (s, t) \<in> ostep E R \<Longrightarrow> s \<succ> t"
  by (force simp: ostep_def subst ctxt elim: ordstep.cases)

inductive
  oKB :: "('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> ('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> bool" (infix "\<turnstile>\<^sub>o\<^sub>K\<^sub>B" 55)
  where
    deduce: "(s, t) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>) \<Longrightarrow> (s, u) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>) \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E \<union> {(t, u)}, R)" |
    orientl: "s \<succ> t \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E - {(s, t)}, R \<union> {(s, t)})" |
    orientr: "t \<succ> s \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E - {(s, t)}, R \<union> {(t, s)})" |
    delete: "(s, s) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E - {(s, s)}, R)" |
    compose: "(t, u) \<in> ostep E (R - {(s, t)}) \<Longrightarrow> (s, t) \<in> R \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E, (R - {(s, t)}) \<union> {(s, u)})" |
    simplifyl: "(s, u) \<in> ostep (E - {(s, t)}) R \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B ((E - {(s, t)}) \<union> {(u, t)}, R)" |
    simplifyr: "(t, u) \<in> ostep (E - {(s, t)}) R \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B ((E - {(s, t)}) \<union> {(s, u)}, R)" |
    collapse: "(t, u) \<in> ostep E (R - {(t, s)}) \<Longrightarrow> (t, s) \<in> R \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E \<union> {(u, s)}, R - {(t, s)})"

lemma oKB_less:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')" and "R \<subseteq> {\<succ>}"
  shows "R' \<subseteq> {\<succ>}"
proof -
  have [dest]: "(x, y) \<in> R \<Longrightarrow> x \<succ> y" for x y using \<open>R \<subseteq> {\<succ>}\<close> by auto
  note assms(1)
  moreover
  { fix s t u assume "(t, u) \<in> rstep (R - {(s, t)})" and "(s, t) \<in> R"
    then have "s \<succ> u" by (induct t u) (blast intro: subst ctxt dest: trans) }
  moreover
  { fix s t u assume "(t, u) \<in> ordstep {\<succ>} (E\<^sup>\<leftrightarrow>)" and "(s, t) \<in> R"
    then have "s \<succ> u" by (cases) (auto dest: trans ctxt) }
  ultimately show ?thesis by (cases) (fastforce simp: ostep_def)+
qed

lemma oKB_rtrancl_less:
  assumes "oKB\<^sup>*\<^sup>* (E, R) (E', R')" and "R \<subseteq> {\<succ>}"
  shows "R' \<subseteq> {\<succ>}"
  using assms by (induct "(E, R)" "(E', R')" arbitrary: E' R') (auto dest!: oKB_less)

lemma oKB_E:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "E \<subseteq> Id \<union> E' \<union> R'\<^sup>\<leftrightarrow> \<union> (ostep E' R') O E' \<union> E' O (ostep E' R')\<inverse>"
  using assms
proof (cases)
  case (simplifyl s u t)
  then have "(s, u) \<in> ostep E' R'"
    by (intro ostep_mono [THEN subsetD, OF _ _ simplifyl(3)]) auto
  then have "(s, t) \<in> (ostep E' R') O E'" using simplifyl by auto
  then show ?thesis using simplifyl by auto
next
  case (simplifyr t u s)
  then have "(t, u) \<in> ostep E' R'"
    by (intro ostep_mono [THEN subsetD, OF _ _ simplifyr(3)]) auto
  then have "(s, t) \<in> E' O (ostep E' R')\<inverse>" using simplifyr by blast
  then show ?thesis using simplifyr by blast
qed auto

lemma oKB_E_rev:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "E' \<subseteq> E \<union> (rstep (R \<union> E\<^sup>\<leftrightarrow>))\<inverse> O (rstep (R \<union> E\<^sup>\<leftrightarrow>)) \<union> (ostep E R)\<inverse> O E \<union> E O (ostep E R)"
  using assms
proof (cases)
  case (simplifyl s u t)
  have "(s, u) \<in> ostep E R"  by (intro ostep_mono [THEN subsetD, OF _ _ simplifyl(3)]) auto
  with simplifyl(4) have "(u, t) \<in> (ostep E R)\<inverse> O E" by auto
  then show ?thesis using simplifyl by auto
next
  case (simplifyr t u s)
  have "(t, u) \<in> ostep E R"
    by (intro ostep_mono [THEN subsetD, OF _ _ simplifyr(3)]) auto
  with simplifyr(4) have "(s, u) \<in> E O (ostep E R)" by auto
  then show ?thesis using simplifyr by auto
next
  case (deduce u s t)
  then have "(s, t) \<in> (rstep (R \<union> E\<^sup>\<leftrightarrow>))\<inverse> O (rstep (R \<union> E\<^sup>\<leftrightarrow>))" by blast
  with deduce show ?thesis by auto
next
  case (collapse t u s)
  have "(t, u) \<in> ostep E R"
    by (intro ostep_mono [THEN subsetD, OF _ _ collapse(3)]) auto
  from ostep_imp_rstep_sym [OF this] subsetD [OF rstep_mono, OF Un_upper1 rstep.rule [OF collapse(4)]]
  have "(u, s) \<in> (rstep (R \<union> E\<^sup>\<leftrightarrow>))\<inverse> O (rstep (R \<union> E\<^sup>\<leftrightarrow>))" by blast
  with collapse show ?thesis by auto
qed auto

lemma oKB_R:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "R \<subseteq> R' \<union> R' O (ostep E' R')\<inverse> \<union> ostep E' R' O E'"
  using assms
proof (cases)
  case (compose t u s)
  then have "(t, u) \<in> ostep E' R'"
    by (intro ostep_mono [THEN subsetD, OF _ _ compose(3)]) auto
  then have "(s, t) \<in> R' O (ostep E' R')\<inverse>" using compose by blast
  then show ?thesis using compose by blast
next
  case (collapse t u s)
  then have "(t, u) \<in> ostep E' R'"
    by (intro ostep_mono [THEN subsetD, OF _ _ collapse(3)]) auto
  then have "(t, s) \<in> ostep E' R' O E'" using collapse by auto
  then show ?thesis using collapse by blast
qed auto

lemma oKB_R_rev:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "R' \<subseteq> R \<union> E\<^sup>\<leftrightarrow> \<union> R O (ostep E R)"
  using assms
proof (cases)
  case (compose s t u)
  have "(s, t) \<in> ostep E R"
    by (intro ostep_mono [THEN subsetD, OF _ _ compose(3)]) auto
  with compose show ?thesis by fast
qed auto

lemma conversion2:
  "(x, y) \<in> r\<^sup>\<leftrightarrow> \<Longrightarrow> (y, z) \<in> r\<^sup>\<leftrightarrow> \<Longrightarrow> (x, z) \<in> r\<^sup>\<leftrightarrow>\<^sup>*"
  by (auto simp: conversion_def intro: rtrancl_trans)

lemma oKB_rstep_subset:
  assumes oKB: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "rstep (E \<union> R) \<subseteq> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using assms
  apply (auto elim!: rstepE dest!: oKB_E [OF oKB, THEN subsetD] oKB_R [OF oKB, THEN subsetD])
     apply (auto  dest!: ostep_imp_rstep_sym)
     apply (rule_tac y = "C\<langle>y \<cdot> \<sigma>\<rangle>" in conversion2; fast)+
  done

lemma oKB_conversion_subset:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "(rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using conversion_mono [OF oKB_rstep_subset [OF assms]] by simp

lemma oKB_rtrancl_conversion_subset:
  assumes "oKB\<^sup>*\<^sup>* (E, R) (E', R')"
  shows "(rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using assms
  by (induct "(E, R)" "(E', R')" arbitrary: E' R')
    (force dest!: oKB_conversion_subset)+

lemma oKB_rstep_subset_rev:
  assumes oKB: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "rstep (E' \<union> R') \<subseteq> (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
  using assms
  apply (auto elim!: rstepE)
   apply (auto dest!: oKB_R_rev [OF oKB, THEN subsetD] oKB_E_rev [OF oKB, THEN subsetD])
     apply (auto dest!: ostep_imp_rstep_sym)
     apply (rule conversion2; fast)+
  done

lemma oKB_conversion_supset:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "(rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
  using conversion_mono [OF oKB_rstep_subset_rev [OF assms]] by simp


section \<open>Abstract ordered completion according to proof for FSCD'17\<close>

inductive_set encstep1 for E :: "('a, 'b) trs" and R :: "('a, 'b) trs"
  where
    estep: "(l, r) \<in> E\<^sup>\<leftrightarrow> \<Longrightarrow> s = C\<langle>l \<cdot> \<sigma>\<rangle> \<Longrightarrow> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<Longrightarrow> l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma> \<Longrightarrow> l \<lhd>\<cdot> s \<Longrightarrow> (s, t) \<in> encstep1 E R"
  | rstep: "(s, t) \<in> rstep R \<Longrightarrow> (s, t) \<in> encstep1 E R"

inductive_set encstep2 for E :: "('a, 'b) trs" and R :: "('a, 'b) trs"
  where
    estep: "(l, r) \<in> E\<^sup>\<leftrightarrow> \<Longrightarrow> s = C\<langle>l \<cdot> \<sigma>\<rangle> \<Longrightarrow> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<Longrightarrow> l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma> \<Longrightarrow> l \<lhd>\<cdot> s \<Longrightarrow> (s, t) \<in> encstep2 E R"
  | rstep: "(l, r) \<in> R \<Longrightarrow> s = C\<langle>l \<cdot> \<sigma>\<rangle> \<Longrightarrow> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<Longrightarrow> l \<lhd>\<cdot> s \<Longrightarrow> (s, t) \<in> encstep2 E R"

inductive
  oKBi :: "('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> ('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> bool" (infix "\<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity>" 55)
  where
    deduce: "(s, t) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>) \<Longrightarrow> (s, u) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>) \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E \<union> {(t, u)}, R)" |
    orientl: "s \<succ> t \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E - {(s, t)}, R \<union> {(s, t)})" |
    orientr: "t \<succ> s \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E - {(s, t)}, R \<union> {(t, s)})" |
    delete: "(s, s) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E - {(s, s)}, R)" |
    compose: "(t, u) \<in> ostep E (R - {(s, t)}) \<Longrightarrow> (s, t) \<in> R \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E, (R - {(s, t)}) \<union> {(s, u)})" |
    simplifyl: "(s, u) \<in> encstep1 (E - {(s, t)}) R \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> ((E - {(s, t)}) \<union> {(u, t)}, R)" |
    simplifyr: "(t, u) \<in> encstep1 (E - {(s, t)}) R \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> ((E - {(s, t)}) \<union> {(s, u)}, R)" |
    collapse: "(t, u) \<in> encstep2 E (R - {(t, s)}) \<Longrightarrow> (t, s) \<in> R \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E \<union> {(u, s)}, R - {(t, s)})"

lemma encstep2_encstep1:
  assumes "(s, t) \<in> encstep2 E R"
  shows "(s, t) \<in> encstep1 E R"
  unfolding encstep1.simps using assms by (cases) auto

lemma encstep1_ostep:
  assumes "(s, t) \<in> encstep1 E R"
  shows "(s, t) \<in> ostep E R"
  using assms
proof (cases)
  case (estep l r C \<sigma>)
  then show ?thesis unfolding ostep_def using ordstep.simps by fast
next
  case rstep
  then show ?thesis unfolding ostep_def rstep_def by fast
qed

lemma encstep2_ostep: "(s, t) \<in> encstep2 E R \<Longrightarrow> (s, t) \<in> ostep E R"
  using encstep2_encstep1 encstep1_ostep by blast

lemma encstep1_mono:
  "E \<subseteq> E' \<Longrightarrow> R \<subseteq> R'\<Longrightarrow> encstep1 E R \<subseteq> encstep1 E' R'"
  using rstep_mono [of R R']
  by (auto elim!: encstep1.cases intro: encstep1.intros)

lemma encstep2_mono:
  "E \<subseteq> E' \<Longrightarrow> R \<subseteq> R'\<Longrightarrow> encstep2 E R \<subseteq> encstep2 E' R'"
  by (auto elim!: encstep2.cases intro: encstep2.intros)

lemma encstep1_imp_rstep_sym:
  "(s, t) \<in> encstep1 E R \<Longrightarrow> (s, t) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>)"
  using encstep1_ostep ostep_imp_rstep_sym by force

lemma encstep2_imp_rstep_sym:
  "(s, t) \<in> encstep2 E R \<Longrightarrow> (s, t) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>)"
  using encstep2_ostep ostep_imp_rstep_sym by force

lemma oKBi_oKB:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  using assms
proof (cases)
  case (simplifyl s u t)
  then show ?thesis using oKB.simplifyl and encstep1_ostep by blast
next
  case (simplifyr t u s)
  then show ?thesis using oKB.simplifyr and encstep1_ostep by blast
next
  case (collapse t u s)
  then show ?thesis using oKB.collapse encstep2_ostep [OF collapse(3)] by blast
qed (blast intro: oKB.intros)+

lemma oKBi_subset:
  assumes oKB: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "E' \<union> R' \<subseteq> (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
  using oKBi_oKB [OF assms] and oKB_rstep_subset_rev subset_rstep [of "E' \<union> R'"] by blast

lemma oKBi_rstep_subset:
  assumes oKB: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "rstep(E' \<union> R') \<subseteq> (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
  using oKBi_oKB [OF assms] and oKB_rstep_subset_rev subset_rstep [of "E' \<union> R'"] by blast

lemma oKBi_conversion_subset:
  assumes oKB: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "(rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
  using oKBi_rstep_subset [OF assms] and conversion_mono and conversion_conversion_idemp by metis

lemma oKBi_E_supset:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "E - E' \<subseteq> (encstep1 E' R' O E'\<^sup>\<leftrightarrow>)\<^sup>\<leftrightarrow> \<union> R'\<^sup>\<leftrightarrow> \<union> Id"
  using assms by (cases) (use encstep1_mono [of "E - {(s, t)}" E' R R' for s t] in blast)+

lemma oKBi_R_supset:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "R - R' \<subseteq> encstep2 E' R' O E' \<union> R' O (ostep E' R')\<inverse>"
  using assms
proof (cases)
  case (compose s t u)
  then have "(s, t) \<in> ostep E' R'"
    using ostep_mono [of E' E "R - {(u, s)}" R'] by auto
  with compose show ?thesis by fast
next
  case (collapse s t u)
  with encstep2_mono [of E "E \<union> {(t, u)}"] show ?thesis by fast
qed auto

lemma oKBi_supset:
  assumes step: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "E \<union> R \<subseteq> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using assms subset_rstep encstep1_imp_rstep_sym encstep2_imp_rstep_sym
  by (smt Diff_partition oKB_rstep_subset oKBi_oKB ordered_completion_axioms sup.boundedE)

lemma oKBi_rstep_supset:
  assumes step: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "rstep(E \<union> R) \<subseteq> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using oKBi_supset [OF assms]
  by (simp add: step oKB_rstep_subset oKBi_oKB ordered_completion_axioms)

lemma oKBi_conversion_supset:
  assumes oKB: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "(rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using oKBi_rstep_supset [OF assms] and conversion_mono and conversion_conversion_idemp by metis

lemma oKBi_conversion:
  assumes oKB: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "(rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using assms and oKBi_conversion_supset and oKBi_conversion_subset by blast

lemma oKBi_less:
  assumes "R \<subseteq> {\<succ>}" and "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "R' \<subseteq> {\<succ>}"
  using assms(1) and oKBi_oKB [OF assms(2)] and oKB_less by auto

lemma oKBi_rtrancl_less:
  assumes "oKBi\<^sup>*\<^sup>* (E, R) (E', R')" and "R \<subseteq> {\<succ>}"
  shows "R' \<subseteq> {\<succ>}"
  using assms oKB_less
proof (induct "(E, R)" "(E', R')" arbitrary: E' R')
  case rtrancl_refl
  then show ?case by fast
next
  case (rtrancl_into_rtrancl S)
  obtain E2 R2 where S: "S = (E2, R2)" by force
  from rtrancl_into_rtrancl show ?case using oKBi_less unfolding S by metis
qed

end

locale okb_irun = ordered_completion +
  fixes R E
  assumes R0: "R 0 = {}"
    and irun: "\<And>i. (E i, R i) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E (Suc i), R (Suc i))"
begin

sublocale kb less enc ..

abbreviation Rinf ("R\<^sub>\<infinity>") where "R\<^sub>\<infinity> \<equiv> (\<Union>i. R i)"
abbreviation Einf ("E\<^sub>\<infinity>") where "E\<^sub>\<infinity> \<equiv> (\<Union>i. E i)"

definition "R\<^sub>\<omega> = (\<Union>i. \<Inter>j\<in>{j. j\<ge>i}. R j)"
definition "E\<^sub>\<omega> = (\<Union>i. \<Inter>j\<in>{j. j\<ge>i}. E j)"

lemma Rw_subset_Rinf: "R\<^sub>\<omega> \<subseteq> R\<^sub>\<infinity>"
  by (auto simp: R\<^sub>\<omega>_def)

lemma Ew_subset_Einf: "E\<^sub>\<omega> \<subseteq> E\<^sub>\<infinity>"
  by (auto simp: E\<^sub>\<omega>_def)

lemma oKBi_rtrancl_i: "oKBi\<^sup>*\<^sup>* (E 0, R 0) (E i, R i)"
  using irun by (induct i) (auto intro: rtranclp.rtrancl_into_rtrancl)

lemma oKBi_conversion_ERi: "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (E i \<union> R i))\<^sup>\<leftrightarrow>\<^sup>*"
  using irun and oKBi_conversion and R0 by (induct i) auto

lemma Ri_less: "R i \<subseteq> {\<succ>}"
  using oKBi_rtrancl_less [OF oKBi_rtrancl_i] and R0 by auto

lemma Rinf_less: "R\<^sub>\<infinity> \<subseteq> {\<succ>}"
  using Ri_less by auto

lemma Rw_less: "R\<^sub>\<omega> \<subseteq> {\<succ>}"
  unfolding R\<^sub>\<omega>_def by (use Ri_less in auto)

lemma Rw_subset_Ri_subset_less_Ew_subset_Ei: "R\<^sub>\<omega> \<subseteq> R\<^sub>\<infinity> \<and> R\<^sub>\<infinity> \<subseteq> {\<succ>} \<and> E\<^sub>\<omega> \<subseteq> E\<^sub>\<infinity>"
  using Rinf_less and Rw_subset_Rinf and Ew_subset_Einf by auto

lemma encstep2D:
  assumes "(s, t) \<in> encstep2 E\<^sub>\<infinity> R\<^sub>\<infinity>"
  shows "\<exists>C \<tau> l r. (l, r) \<in> R\<^sub>\<infinity> \<union> E\<^sub>\<infinity>\<^sup>\<leftrightarrow> \<and> s = C\<langle>l \<cdot> \<tau>\<rangle> \<and> t = C\<langle>r \<cdot> \<tau>\<rangle> \<and> l \<cdot> \<tau> \<succ> r \<cdot> \<tau> \<and> l \<lhd>\<cdot> s"
  using assms
proof (cases)
  case (estep l r C \<sigma>)
  then show ?thesis by blast
next
  case (rstep l r C \<sigma>)
  from rstep(1) and Ri_less and subst [of l r] have "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma>" by auto
  with rstep show ?thesis by blast
qed

definition lessencp (infix "\<cdot>\<succ>" 50) where "lessencp s t \<longleftrightarrow> (s, t) \<in> (relto ({\<cdot>\<rhd>} \<union> {\<succ>}) {\<cdot>\<unrhd>})\<^sup>+"
abbreviation lessenc ("{\<cdot>\<succ>}") where "lessenc \<equiv> {(s, t). s \<cdot>\<succ> t}"

definition leqencp (infix "\<cdot>\<succeq>" 50) where "leqencp s t \<longleftrightarrow> (s, t) \<in> (relto ({\<cdot>\<rhd>} \<union> {\<succ>}) {\<cdot>\<unrhd>})\<^sup>+ \<union> {\<cdot>\<unrhd>}\<^sup>*"
abbreviation leqenc ("{\<cdot>\<succeq>}") where "leqenc \<equiv> {(s, t). s \<cdot>\<succeq> t}"

lemma SN_lessenc: "SN {\<cdot>\<succ>}"
  using reduction_order.SN_encomp_Un_less_relto_encompeq
    SN_encomp_Un_less_relto_encompeq SN_trancl_SN_conv unfolding lessencp_def by auto

lemma lessenc_compat1: "{\<cdot>\<succ>} O {\<cdot>\<unrhd>} \<subseteq> {\<cdot>\<succ>}"
proof
  fix s t
  let ?R = "{\<cdot>\<unrhd>}\<^sup>* O ({\<cdot>\<rhd>} \<union> {\<succ>}) O {\<cdot>\<unrhd>}\<^sup>*"
  assume "(s, t) \<in> {\<cdot>\<succ>} O {\<cdot>\<unrhd>}"
  then obtain u where u: "(u, t) \<in> {\<cdot>\<unrhd>}" "s \<cdot>\<succ> u" by auto
  then obtain v where v: "(v, u) \<in> ?R" "(s, v) \<in> ?R\<^sup>*" using tranclD2 unfolding lessencp_def by metis
  from u(1) v(1) have "(v, t) \<in> ?R O {\<cdot>\<unrhd>}" by blast
  with u(1) v(1) have "(v, t) \<in> ?R" using O_assoc O_mono2 [of "{\<cdot>\<unrhd>} O {\<cdot>\<unrhd>}\<^sup>*" "{\<cdot>\<unrhd>}\<^sup>*"]
    by (smt r_into_rtrancl relcomp.relcompI rtrancl_idemp_self_comp u(1) v(1))
  with v(2) have "s \<cdot>\<succ> t" unfolding trancl_unfold_right lessencp_def by blast
  then show "(s, t) \<in> {\<cdot>\<succ>}" by auto
qed

lemma leqenc_compat1: "{\<cdot>\<succeq>} O {\<cdot>\<unrhd>} \<subseteq> {\<cdot>\<succeq>}"
proof
  fix s t
  let ?R = "({\<cdot>\<unrhd>}\<^sup>* O ({\<cdot>\<rhd>} \<union> {\<succ>}) O {\<cdot>\<unrhd>}\<^sup>*) \<union> {\<cdot>\<unrhd>}\<^sup>*"
  assume "(s, t) \<in> {\<cdot>\<succeq>} O {\<cdot>\<unrhd>}"
  then obtain u where u: "s \<cdot>\<succeq> u" "(u, t) \<in> {\<cdot>\<unrhd>}" by auto
  then consider "s \<cdot>\<succ> u" | "(s, u) \<in> {\<cdot>\<unrhd>}\<^sup>*"
    unfolding leqencp_def lessencp_def Un_iff by blast
  then show "(s, t) \<in> {\<cdot>\<succeq>}"
  proof (cases)
    case 1
    with lessenc_compat1 u(2) show ?thesis unfolding leqencp_def lessencp_def by auto
  next
    case 2
    with u(2) rtrancl_comp_trancl_conv [of "{\<cdot>\<unrhd>}"] have "(s, t) \<in> {\<cdot>\<unrhd>}\<^sup>*" by auto
    then show ?thesis unfolding leqencp_def Un_iff mem_Collect_eq split by auto
  qed
qed

lemma lessleqenc_compat1: "{\<cdot>\<succ>} O {\<cdot>\<succeq>} \<subseteq> {\<cdot>\<succ>}"
proof
  fix x z
  assume "(x, z) \<in> {\<cdot>\<succ>} O {\<cdot>\<succeq>}"
  then obtain y where y: "(x, y) \<in> {\<cdot>\<succ>}" "(y, z) \<in> {\<cdot>\<succeq>}" by auto
  then consider "y \<cdot>\<succ> z" | "(y, z) \<in> {\<cdot>\<unrhd>}\<^sup>*"
    unfolding leqencp_def lessencp_def Un_iff by blast
  then show "(x, z) \<in> {\<cdot>\<succ>}"
  proof (cases)
    case 1
    with y(1) show ?thesis using trancl_trans unfolding lessencp_def by auto
  next
    case 2
    from encompeq_trans trans_inv_image transI have t: "trans {\<cdot>\<unrhd>}"
      by (smt Pair_inject case_prodE case_prodI2 mem_Collect_eq)
    have "{\<cdot>\<unrhd>}\<^sup>* = {\<cdot>\<unrhd>}" by(rule trans_refl_imp_rtrancl_id, insert t encompeq_refl refl_O_iff, auto)
    with 2 [unfolded this] y(1) lessenc_compat1 show ?thesis by auto
  qed
qed

lemma lessenc_compat2: "{\<cdot>\<unrhd>} O {\<cdot>\<succ>} \<subseteq> {\<cdot>\<succ>}"
proof
  fix s t
  let ?R = "{\<cdot>\<unrhd>}\<^sup>* O ({\<cdot>\<rhd>} \<union> {\<succ>}) O {\<cdot>\<unrhd>}\<^sup>*"
  assume "(s, t) \<in> {\<cdot>\<unrhd>} O {\<cdot>\<succ>}"
  then obtain u where u: "(s, u) \<in> {\<cdot>\<unrhd>}" "u \<cdot>\<succ> t" by auto
  then obtain v where v: "(u, v) \<in> ?R" "(v, t) \<in> ?R\<^sup>*" using tranclD unfolding lessencp_def by metis
  from u(1) v(1) have "(s, v) \<in> {\<cdot>\<unrhd>} O ?R" by blast
  with u(1) v(1) have "(s, v) \<in> ?R" using O_assoc O_mono2 [of "{\<cdot>\<unrhd>} O {\<cdot>\<unrhd>}\<^sup>*" "{\<cdot>\<unrhd>}\<^sup>*"]
    by (smt r_into_rtrancl relcomp.relcompI rtrancl_idemp_self_comp u(1) v(1))
  with v(2) have "s \<cdot>\<succ> t" unfolding trancl_unfold_left lessencp_def by blast
  then show "(s, t) \<in> {\<cdot>\<succ>}" by auto
qed

lemma encstep1_less:
  assumes "(s, t) \<in> encstep1 (E i) (R i)"
  shows "s \<succ> t"
  using assms
proof (cases)
  case (estep l r C \<sigma>)
  with ctxt [OF estep(4)] show ?thesis unfolding lessencp_def by fast
next
  case rstep
  with compatible_rstep_imp_less [OF Ri_less] show ?thesis unfolding lessencp_def by blast
qed

lemma encstep1_lessenc:
  assumes "(s, t) \<in> encstep1 (E i) (R i)"
  shows "s \<cdot>\<succ> t"
  using assms encstep1_less unfolding lessencp_def by fast

lemma trans_less_enc: "trans {\<cdot>\<succ>}"
proof -
  from commutes_rewrel_encomp [of "{\<succ>}"] and ctxt and subst have "{\<cdot>\<unrhd>} O {\<succ>} \<subseteq>  {\<succ>} O {\<cdot>\<unrhd>}" by simp
  with trans_relto [of "{\<succ>}" "{\<cdot>\<unrhd>}"] and trans show ?thesis unfolding lessencp_def by auto
qed

lemma refl_encompeq: "refl {\<cdot>\<unrhd>}"
  using encompeq_refl reflpI reflp_refl_eq by fastforce

abbreviation "mulless \<equiv> s_mul_ext {\<cdot>\<unrhd>} {\<cdot>\<succ>}"
abbreviation "mulleq \<equiv> ns_mul_ext {\<cdot>\<unrhd>} {\<cdot>\<succ>}"

sublocale lessenc_op: SN_order_pair mulless "ns_mul_ext {\<cdot>\<unrhd>} {\<cdot>\<succ>}"
proof (intro SN_order_pair.mul_ext_SN_order_pair)
  show "SN_order_pair {\<cdot>\<succ>} {\<cdot>\<unrhd>}"
  proof
    show "trans {\<cdot>\<succ>}" using trans_less_enc by auto
  next
    show "refl {\<cdot>\<unrhd>}" using refl_encompeq .
  next
    show "trans {\<cdot>\<unrhd>}" using encomp_order.order_trans transpI transp_trans by metis
  next
    from SN_lessenc show "SN {\<cdot>\<succ>}" by auto
  next
    from lessenc_compat2 show "{\<cdot>\<unrhd>} O {\<cdot>\<succ>} \<subseteq> {\<cdot>\<succ>}" by auto
  next
    from lessenc_compat1 show "{\<cdot>\<succ>} O {\<cdot>\<unrhd>} \<subseteq> {\<cdot>\<succ>}" by auto
  qed
qed

lemma mset_two:
  assumes "s \<cdot>\<succ> u"
  shows "({#s, t#}, {#u, t#}) \<in> mulless"
proof -
  from assms s_mul_ext_singleton have su: "({#s#}, {#u#}) \<in> mulless" by auto
  from ns_mul_ext_refl [OF refl_encompeq] have "({#t#}, {#t#}) \<in> mulleq" by auto
  from s_ns_mul_ext_union_compat [OF su this] show ?thesis
    using add.commute add_mset_add_single by metis
qed

lemma mset_two2:
  assumes "s \<cdot>\<succ> u"
  shows "({#t, s#}, {#t, u#}) \<in> mulless"
  using mset_two [OF assms] add_mset_commute by (metis (no_types, lifting))

lemma rstep_imp_mstep:
  assumes "(t, u) \<in> rstep RR" and "\<exists>t' \<in># S. t' \<succeq> t" and "\<exists>u' \<in># S. u' \<succeq> u"
  shows "(t, u) \<in> mstep S RR"
  using assms by (auto simp: mstep_iff dest: trans)

lemma mstep_succeq_mono:
  assumes "\<forall>m \<in># M. \<exists>n \<in># N. n \<succeq> m"
  shows "mstep M RR \<subseteq> mstep N RR"
proof
  fix s t
  assume mstep: "(s, t) \<in> mstep M RR"
  with assms trans show "(s, t) \<in> mstep N RR" unfolding mstep_def mem_Collect_eq split
    by (smt sup2E sup2I1)
qed

lemma msteps_succeq_mono:
  assumes "\<forall>m \<in># M. \<exists>n \<in># N. n \<succeq> m"
  shows "(mstep M RR)\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (mstep N RR)\<^sup>\<leftrightarrow>\<^sup>*"
  by (rule mstep_succeq_mono [OF assms, THEN conversion_mono])

lemma mstep_ctxt:
  assumes "(s, t) \<in> mstep {#u, v#} RR"
  shows "(C\<langle>s \<cdot> \<sigma>\<rangle>, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> mstep {#C\<langle>u \<cdot> \<sigma>\<rangle>, C\<langle>v \<cdot> \<sigma>\<rangle>#} RR"
proof -
  from assms have step: "(C\<langle>s \<cdot> \<sigma>\<rangle>, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> rstep RR" unfolding mstep_def by auto
  from assms have "u \<succeq> s \<or> v \<succeq> s" "u \<succeq> t \<or> v \<succeq> t" unfolding mstep_def by auto
  with subst ctxt have "C\<langle>u \<cdot> \<sigma>\<rangle> \<succeq> C\<langle>s \<cdot> \<sigma>\<rangle> \<or> C\<langle>v \<cdot> \<sigma>\<rangle> \<succeq> C\<langle>s \<cdot> \<sigma>\<rangle>" "C\<langle>u \<cdot> \<sigma>\<rangle> \<succeq> C\<langle>t \<cdot> \<sigma>\<rangle> \<or> C\<langle>v \<cdot> \<sigma>\<rangle> \<succeq> C\<langle>t \<cdot> \<sigma>\<rangle>"
    unfolding mstep_def by blast+
  with step show ?thesis unfolding mstep_def by fastforce
qed

lemma msteps_ctxt:
  assumes "(s, t) \<in> (mstep {#u, v#} RR)\<^sup>\<leftrightarrow>\<^sup>*"
  shows "(C\<langle>s \<cdot> \<sigma>\<rangle>, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> (mstep {#C\<langle>u \<cdot> \<sigma>\<rangle>, C\<langle>v \<cdot> \<sigma>\<rangle>#} RR)\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  from mstep_ctxt have x: "\<And> s t u v. (s, t) \<in> (mstep {#u, v#} RR)\<^sup>\<leftrightarrow> \<Longrightarrow> (C\<langle>s \<cdot> \<sigma>\<rangle>, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> (mstep {#C\<langle>u \<cdot> \<sigma>\<rangle>, C\<langle>v \<cdot> \<sigma>\<rangle>#} RR)\<^sup>\<leftrightarrow>" by blast
  from assms rtrancl_imp_relpow obtain n where "(s, t) \<in> (mstep {#u, v#} RR)\<^sup>\<leftrightarrow> ^^ n" by fast
  then have "(C\<langle>s \<cdot> \<sigma>\<rangle>, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> (mstep {#C\<langle>u \<cdot> \<sigma>\<rangle>, C\<langle>v \<cdot> \<sigma>\<rangle>#} RR)\<^sup>\<leftrightarrow> ^^ n"
  proof (induct n arbitrary: s)
    case (Suc n)
    then obtain s' where s': "(s, s') \<in> (mstep {#u, v#} RR)\<^sup>\<leftrightarrow>" "(s', t) \<in> (mstep {#u, v#} RR)\<^sup>\<leftrightarrow> ^^ n"
      unfolding relpow_Suc by blast
    from x [OF s'(1)] Suc(1) [OF s'(2)] show ?case unfolding relpow_Suc by blast
  qed simp
  with relpow_imp_rtrancl show ?thesis by blast
qed

lemma Ei_subset_EwRi: "mstep S E\<^sub>\<infinity> \<subseteq> (mstep S (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  { fix t0 u0
    assume "(t0, u0) \<in> rstep E\<^sub>\<infinity>" and St: "\<exists>s \<in># S. s \<succeq> t0" and Su: "\<exists>s \<in># S. s \<succeq> u0"
    with rstep_imp_C_s_r [OF this(1)] obtain C \<sigma> t u
      where Cs: "t0 = C\<langle>t \<cdot> \<sigma>\<rangle>" "u0 = C\<langle>u \<cdot> \<sigma>\<rangle>" and tu: "(t, u) \<in> E\<^sub>\<infinity>" by blast
    from tu St Su have "(C\<langle>t \<cdot> \<sigma>\<rangle>, C\<langle>u \<cdot> \<sigma>\<rangle>) \<in> (mstep S (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding Cs
    proof (induct "{#t, u#}" arbitrary: t u S C \<sigma> rule: SN_induct [OF lessenc_op.SN])
      case 1
      note sc = this
      let ?t = "C\<langle>t \<cdot> \<sigma>\<rangle>" and ?u = "C\<langle>u \<cdot> \<sigma>\<rangle>"
      have comm: "\<And>t u. {#t, u#} =  {#u, t#}" using add_mset_commute by auto
      note r_imp_msteps = rstep_imp_mstep [OF _ sc(4) sc(3)] rstep_imp_mstep [OF _ sc(3) sc(4)]
      show ?case
      proof (cases "(?t, ?u) \<in> rstep E\<^sub>\<omega>")
        case True
        with r_imp_msteps mstep_Un [of _ "R\<^sub>\<infinity>" "E\<^sub>\<omega>"] show ?thesis by auto
      next
        case False
        with rstep_imp_C_s_r have "(t, u) \<notin> E\<^sub>\<omega>" by fast
        with sc(2) obtain j where "(t, u) \<in> E j" and "\<exists>i>j. (t, u) \<notin> E i"
          unfolding E\<^sub>\<omega>_def le_eq_less_or_eq by blast
        moreover define i where "i = (LEAST i. i > j \<and> (t, u) \<notin> E i)"
        ultimately have "i > j" and not_i: "(t, u) \<notin> E i" by (metis (lifting) LeastI)+
        then have "i - 1 < i" and [simp]: "Suc (i - Suc 0) = i" by auto
        from not_less_Least [OF this(1) [unfolded i_def], folded i_def] and \<open>i > j\<close> and \<open>(t, u) \<in> E j\<close>
        have pred_i: "(t, u) \<in> E (i - 1)" by (cases "i = Suc j") auto
        have i_simp [simp]: "Suc (i - 1) = i" by auto
        let ?enc = "encstep1 (E (Suc (i - 1))) (R (Suc (i - 1))) O (E (Suc (i - 1)))\<^sup>\<leftrightarrow>"
        from pred_i oKBi_E_supset [OF irun [of "i - 1"]] and not_i
        consider "(t, u) \<in> (R (Suc (i - 1)))\<^sup>\<leftrightarrow>" | "u = t" | "(t, u) \<in> ?enc\<^sup>\<leftrightarrow>" by auto blast
        then show ?thesis
        proof (cases)
          case 1
          then have "(?t, ?u) \<in> rstep (R\<^sub>\<infinity>\<^sup>\<leftrightarrow>)" by auto
          with r_imp_msteps [of "R\<^sub>\<infinity>"] converse_iff have "(?t, ?u) \<in> (mstep S R\<^sub>\<infinity>) \<or> (?u, ?t) \<in> (mstep S R\<^sub>\<infinity>)"
            by fastforce
          with mstep_Un conversion_inv show ?thesis by blast
        next
          case 3
          then have "(t, u) \<in> ?enc \<or> (u, t) \<in> ?enc" by auto
          with sc obtain t' u' where tu': "(t', u') \<in> ?enc" "(t=t' \<and> u=u') \<or> (t=u' \<and> u=t')" by blast
          then have tu_tu': "{# t, u #} = {# t', u' #}" by force
          from sc(3) sc(4) tu'(2) trans have Stu: "\<exists>s \<in># S. s \<succeq> C\<langle>u' \<cdot> \<sigma>\<rangle> \<and> (\<exists>s \<in># S. s \<succeq> C\<langle>t' \<cdot> \<sigma>\<rangle>)" by meson
          from tu'(1) obtain v where v: "(t', v) \<in> encstep1 (E i) (R i)" "(v, u') \<in> (E i)\<^sup>\<leftrightarrow>"
            unfolding i_simp by auto
          from tu_tu' mset_two [OF encstep1_lessenc [OF v(1)]]
          have dec: "({# t, u #}, {# v, u' #}) \<in> mulless" by auto
          then have dec': "({# t, u #}, {# u', v #}) \<in> mulless" using add_mset_commute by metis
          from encstep1_ostep [OF v(1), THEN ostep_imp_less [OF Ri_less]] have tv: "t' \<succeq> v" by blast
          have vu: "(C\<langle>v\<cdot> \<sigma>\<rangle>, C\<langle>u'\<cdot> \<sigma>\<rangle>) \<in> (mstep S (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
          proof -
            from trans subst ctxt tv Stu have Sv: "\<exists>s \<in># S. s \<succeq> C\<langle>v\<cdot> \<sigma>\<rangle>" by fast
            from v(2) have "(v, u') \<in> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>" by auto
            with sc(1) [OF dec] sc(1) [OF dec'] Sv Stu show ?thesis
              unfolding Un_iff conversion_inv by blast
          qed
          from v(1) have "(C\<langle>t' \<cdot> \<sigma>\<rangle>, C\<langle>v \<cdot> \<sigma>\<rangle>) \<in> (mstep S (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
          proof (cases)
            case (estep l r D \<tau>)
            then have lr: "(l, r) \<in> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>" by auto
            with encompeq.intros [of t' D "l \<cdot> \<tau>" Var] estep(2) have tl: "l \<cdot> \<tau> \<unlhd>\<cdot> t'" by auto
            with encompeq.intros [of "r \<cdot> \<tau>" \<box>] estep(2) have "r \<unlhd>\<cdot> r \<cdot> \<tau>" by auto
            with estep(4) tl have tr: "t' \<cdot>\<succ> r" unfolding lessencp_def by blast
            from estep(5) have "t' \<cdot>\<succ> l" unfolding lessencp_def by blast
            with tr ns_s_mul_ext_union_multiset_l [OF ns_mul_ext_bottom [of "{#u'#}"], of _ "{#l, r#}"]
            have "({#u', t'#}, {#l, r#}) \<in> mulless" using multi_member_last by auto
            with tu_tu' have dec: "({#t, u#}, {#l, r#}) \<in> mulless" using comm by metis
            with comm [of l] have dec': "({#t, u#}, {#r, l#}) \<in> mulless" by presburger
            note ih = sc(1) [OF dec, of "{#l, r#}" \<box> Var] sc(1) [OF dec', of "{#l, r#}" \<box> Var]
            from lr [unfolded Un_iff] ih have "(l, r) \<in> (mstep {#l, r#} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
              unfolding conversion_inv [of r] ctxt.cop_nil subst_apply_term_empty by force
            from msteps_ctxt [OF this] estep(2) estep(3)
            have mstep: "(t', v) \<in> (mstep {#t', v#} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
            from Stu tv trans ctxt [OF subst] have "\<exists>s\<in>#S. s \<succeq> C\<langle>v \<cdot> \<sigma>\<rangle>" by fast
            with Stu tv have "\<forall>m\<in>#{#C\<langle>t' \<cdot> \<sigma>\<rangle>, C\<langle>v \<cdot> \<sigma>\<rangle>#}. \<exists>n\<in>#S. n \<succeq> m" by auto
            from msteps_ctxt [OF mstep] conversion_mono [OF mstep_succeq_mono [OF this]]
            show ?thesis by blast
          next
            case rstep
            with rstep_subset_less [OF Ri_less] have tv': "t' \<succ> v" by blast
            from rstep have "(t', v) \<in> rstep (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>)" by fast
            from rstep_imp_mstep [OF this] tv'  have "(t', v) \<in> mstep {#t', t'#} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>)" by auto
            from mstep_ctxt [OF this, of C \<sigma>] mstep_succeq_mono [of "{#C\<langle>t' \<cdot> \<sigma>\<rangle>, C\<langle>t' \<cdot> \<sigma>\<rangle>#}" S] Stu
            show ?thesis by fastforce
          qed
          from transD [OF conversion_trans, OF this vu] tu'(2) symD [OF conversion_sym]
          show ?thesis by metis
        qed auto
      qed
    qed
    with Cs have "(t0, u0) \<in> (mstep S (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  }
  then show ?thesis unfolding mstep_def by blast
qed

lemma Einf_sym_mstep: "mstep S (E\<^sub>\<infinity>\<^sup>\<leftrightarrow>) \<subseteq> (mstep S (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  have inv: "mstep S (E\<^sub>\<infinity>\<inverse>) = (mstep S E\<^sub>\<infinity>)\<inverse>"
    unfolding mstep_def reverseTrs [symmetric] converse_iff by auto
  from Ei_subset_EwRi show ?thesis unfolding mstep_Un inv Un_subset_iff
    by (metis converse_mono conversion_converse)
qed

abbreviation "lexless \<equiv> lex_two {\<cdot>\<succ>} Id {\<cdot>\<succ>}"

lemma SN_lexless: "SN lexless"
  using SN_lessenc
  by (intro lex_two) (auto simp only: SN_trancl_SN_conv intro: SN_encomp_Un_less_relto_encompeq)

lemma Ri_subset_ERw: "mstep S R\<^sub>\<infinity> \<subseteq> (mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  { fix t u
    assume "(t, u) \<in> rstep R\<^sub>\<infinity>" and St: "\<exists>s \<in># S. s \<succeq> t" and Su: "\<exists>s \<in># S. s \<succeq> u"
    with rstep_imp_C_s_r [OF this(1)] obtain C \<sigma> l r
      where Cs: "t = C\<langle>l \<cdot> \<sigma>\<rangle>" "u = C\<langle>r \<cdot> \<sigma>\<rangle>" and tu: "(l, r) \<in> R\<^sub>\<infinity>" by blast
    from tu St Su have "(C\<langle>l \<cdot> \<sigma>\<rangle>, C\<langle>r \<cdot> \<sigma>\<rangle>) \<in> (mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding Cs
    proof (induct "(l, r)" arbitrary: l r S C \<sigma> rule: SN_induct [OF SN_lexless])
      case 1
      note sc = this
      with Ri_less have lr: "l \<succ> r" by auto
      let ?t = "C\<langle>l \<cdot> \<sigma>\<rangle>" and ?u = "C\<langle>r \<cdot> \<sigma>\<rangle>"
      show ?case
      proof (cases "(l, r) \<in> R\<^sub>\<omega>")
        case True
          (* rule l \<rightarrow> r is persistent *)
        then have "(?t, ?u) \<in> rstep R\<^sub>\<omega>" by auto
        with rstep_imp_mstep [OF this sc(3) sc(4)] mstep_Un show ?thesis by auto
      next
        case False
          (* rule l \<rightarrow> r was removed in step (E i, R i) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E (i+1), R (i+1)) *)
        with sc(2) obtain j where "(l, r) \<in> R j" and "\<exists>i>j. (l, r) \<notin> R i"
          unfolding R\<^sub>\<omega>_def le_eq_less_or_eq by blast
        moreover define i where "i = (LEAST i. i > j \<and> (l, r) \<notin> R i)"
        ultimately have "i > j" and not_i: "(l, r) \<notin> R i" by (metis (lifting) LeastI)+
        then have "i - 1 < i" and [simp]: "Suc (i - Suc 0) = i" by auto
        from not_less_Least [OF this(1) [unfolded i_def], folded i_def] and \<open>i > j\<close> and \<open>(l, r) \<in> R j\<close>
        have pred_i: "(l, r) \<in> R (i - 1)" by (cases "i = Suc j") auto
        have i_simp [simp]: "Suc (i - 1) = i" by auto
            (* first verify that (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>) conversion below l corresponds to a (R\<^sub>\<omega> \<union> E\<^sub>\<omega>) conversion *)
        { fix S'
          assume below_l: "\<forall> s' \<in># S'. l \<cdot>\<succ> s'"
          { fix s t
            assume "(s, t) \<in> mstep S' R\<^sub>\<infinity>"
            then have mstep: "(s, t) \<in> rstep R\<^sub>\<infinity>" "\<exists>s' \<in># S'. s' \<succeq> s" "\<exists>s' \<in># S'. s' \<succeq> t"
              unfolding mstep_def by auto
            then obtain C \<sigma> l1 r1 where l1r1: "s = C\<langle>l1 \<cdot> \<sigma>\<rangle>" "t = C\<langle>r1 \<cdot> \<sigma>\<rangle>" "(l1, r1) \<in> R\<^sub>\<infinity>" by fast
            from mstep obtain s' where s': "s' \<in># S'" "s' \<cdot>\<succeq> s" unfolding leqencp_def by blast
            from l1r1 encompeq.intros have sl1: "s \<cdot>\<unrhd> l1" by auto
            from s' below_l have ls': "l \<cdot>\<succ> s'" by auto
            with s' lessleqenc_compat1 have "l \<cdot>\<succ> s" by auto
            with sl1 lessenc_compat1 have "l \<cdot>\<succ> l1" by auto
            then have "((l, r), (l1, r1)) \<in> lexless" by force
            from sc(1) [OF this l1r1(3)] mstep have "(s, t) \<in> (mstep S' (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
              unfolding l1r1 by meson
          } note step = this
          from r_into_rtrancl have "mstep S' E\<^sub>\<omega> \<subseteq> (mstep S' E\<^sub>\<omega>)\<^sup>\<leftrightarrow>\<^sup>*" by auto
          with step mstep_Un have "mstep S' (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>) \<subseteq> (mstep S' (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
          from conversion_mono [OF this] have R: "(mstep S' (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (mstep S' (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
            by simp
        } note mstep_Rinf_msteps_ERw = this
          (* derive equation (1): (R\<^sub>\<infinity> \<union> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>) steps below l correspond to a (R\<^sub>\<omega> \<union> E\<^sub>\<omega>) conversion *)
        { fix S'
          assume below_l: "\<forall> s' \<in># S'. l \<cdot>\<succ> s'"
          note R = mstep_Rinf_msteps_ERw [OF this]
          with Einf_sym_mstep have "mstep S' (E\<^sub>\<infinity>\<^sup>\<leftrightarrow>) \<subseteq> (mstep S' (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
          with R mstep_Un have "mstep S' (R\<^sub>\<infinity> \<union> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>) \<subseteq> (mstep S' (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
          from conversion_mono [OF this] have "(mstep S' (R\<^sub>\<infinity> \<union> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (mstep S' (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
            by simp
        } note msteps_ERinf_mstep_Rw = this
          (* case distinction according to Lemma 31 *)
        from pred_i oKBi_R_supset [OF irun [of "i - 1"]] and not_i
        consider "(l, r) \<in> encstep2 (E i) (R i) O E i" | "(l, r) \<in> R i O (ostep (E i) (R i))\<inverse>"
          unfolding i_simp by blast
        then have "(l, r) \<in> (mstep {#l, r#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
        proof (cases)
          (* collapse step *)
          case 1
          then obtain u where u: "(l, u) \<in> encstep2 (E i) (R i)" "(u, r) \<in> E i" by auto
              (* l \<longleftrightarrow>\<^sub>R\<^sub>\<omega> \<^sub>\<union> \<^sub>E\<^sub>\<omega>\<^sup>{\<^sup>l\<^sup>, \<^sup>r\<^sup>}\<^sup>* u *)
          with encstep2_mono [of "E i" "E\<^sub>\<infinity>" "R i" "R\<^sub>\<infinity>"] have lu: "(l, u) \<in> encstep2 E\<^sub>\<infinity> R\<^sub>\<infinity>" by auto
          from encstep2D [OF this] obtain D \<tau> l0 r0 where
            D: "(l0, r0) \<in> R\<^sub>\<infinity> \<union> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>" "l = D\<langle>l0 \<cdot> \<tau>\<rangle>" "u = D\<langle>r0 \<cdot> \<tau>\<rangle>" " l0 \<cdot> \<tau> \<succ> r0 \<cdot> \<tau>" "l0 \<lhd>\<cdot> l" by blast
          then have l_l0: "l \<cdot>\<succ> l0" unfolding lessencp_def by auto
          from D encompeq.intros [of l D _ Var] have l_l0\<tau>: "l \<cdot>\<unrhd> l0 \<cdot> \<tau>" by auto
          from encompeq.intros [of _ \<box>] have "r0 \<cdot> \<tau> \<cdot>\<unrhd> r0" by auto
          with l_l0\<tau> D(4) have l_r0: "l \<cdot>\<succ> r0" unfolding lessencp_def by blast
          have "(l0, r0) \<in> (mstep {#l0, r0#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
          proof (cases "(l0, r0) \<in> R\<^sub>\<infinity>")
            case True
            with Ri_less have l_gt_r: "l0 \<succ> r0" by auto
            from l_l0 have "((l, r), (l0, r0)) \<in> lexless" by auto
            from sc(1) [OF this True, of "{#l0, r0#}" \<box> Var, simplified] l_gt_r show ?thesis by auto
          next
            case False
            with D(1) have "(l0, r0) \<in> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>" by auto
            then have "(l0, r0) \<in> mstep {#l0, r0#} (E\<^sub>\<infinity>\<^sup>\<leftrightarrow>)" unfolding mstep_def by auto
            with in_mono [OF Einf_sym_mstep] have l0r0: "(l0, r0) \<in> (mstep {#l0, r0#} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
            from l_l0 l_r0 have "\<forall> s' \<in># {#l0, r0#}. l \<cdot>\<succ> s'" by auto
            from l0r0 mstep_Un mstep_Rinf_msteps_ERw [OF this] show ?thesis by auto
          qed
          from D msteps_ctxt [OF this] have lu: "(l, u) \<in> (mstep {#l, u#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
          from u(1) encstep1_less [OF encstep2_encstep1] have l_succeq_u: "l \<succeq> u" by auto
          then have "\<forall>m\<in>#{#l, u#}. \<exists>n\<in>#{#l, r#}. n \<succeq> m" by auto
          from lu msteps_succeq_mono [OF this] have lu: "(l, u) \<in> (mstep {#l, r#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by blast
              (* u \<longleftrightarrow>\<^sub>R\<^sub>\<omega> \<^sub>\<union> \<^sub>E\<^sub>\<omega>\<^sup>{\<^sup>l\<^sup>, \<^sup>r\<^sup>}\<^sup>* r *)
          from u(1) encstep1_lessenc [OF encstep2_encstep1] have "l \<cdot>\<succ> u" by auto
          with lr have lt_l: "\<forall> s' \<in># {#u, r#}. l \<cdot>\<succ> s'" unfolding lessencp_def by force
          from u(2) have "(u, r) \<in> mstep {#u, r#} (R\<^sub>\<infinity> \<union> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>)" unfolding mstep_def by auto
          from r_into_rtrancl [OF this] msteps_ERinf_mstep_Rw [OF lt_l]
          have ur: "(u, r) \<in> (mstep {#u, r#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
          from l_succeq_u have "\<forall>m\<in>#{#u, r#}. \<exists>n\<in>#{#l, r#}. n \<succeq> m" by auto
          from ur msteps_succeq_mono [OF this] have "(u, r) \<in> (mstep {#l, r#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by blast
          with lu transD [OF conversion_trans] show ?thesis by auto
        next
          (* compose step *)
          case 2
          then obtain u where u: "(l, u) \<in> R i" "(u, r) \<in> (ostep (E i) (R i))\<inverse>" by auto
              (* l \<longleftrightarrow>\<^sub>R\<^sub>\<omega> \<^sub>\<union> \<^sub>E\<^sub>\<omega>\<^sup>{\<^sup>l\<^sup>, \<^sup>r\<^sup>}\<^sup>* u*)
          with ostep_imp_less [OF Ri_less] have ru: "r \<succ> u" by auto
          then have "r \<cdot>\<succ> u" unfolding lessencp_def by blast
          then have "((l, r), (l, u)) \<in> lexless" unfolding lex_two.simps mem_Collect_eq split by auto
          from trans [OF lr ru] sc(1) [OF this, of "{#l, r#}" \<box> Var, simplified] u(1)
          have lu: "(l, u) \<in> (mstep {#l, r#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
              (* u \<longleftrightarrow>\<^sub>R\<^sub>\<omega> \<^sub>\<union> \<^sub>E\<^sub>\<omega>\<^sup>{\<^sup>l\<^sup>, \<^sup>r\<^sup>}\<^sup>* r*)
          from ostep_imp_rstep_sym u(2) have "(r, u) \<in> rstep (R i \<union> (E i)\<^sup>\<leftrightarrow>)" by auto
          with rstep_UN rstep_union have rstep: "(r, u) \<in> rstep (R\<^sub>\<infinity> \<union> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>)" by fast
          from rstep have mstep: "(r, u) \<in> mstep {#u, r#} (R\<^sub>\<infinity> \<union> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>)"
            unfolding mstep_def mem_Collect_eq split by auto
          from lr trans [OF lr ru] have lt_l: "\<forall> s' \<in># {#u, r#}. l \<cdot>\<succ> s'"
            unfolding lessencp_def by force
          from msteps_ERinf_mstep_Rw [OF this] r_into_rtrancl [OF mstep]
          have ur: "(u, r) \<in> (mstep {#u, r#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding conversion_inv [of u] by auto
          have "(u, r) \<in> (mstep {#l, r#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
            by(rule subsetD [OF msteps_succeq_mono], insert trans [OF lr ru] ur, auto)
          with lu transD [OF conversion_trans] show ?thesis by auto
        qed
        from msteps_ctxt [OF this] have tu: "(?t, ?u) \<in> (mstep {#?t, ?u#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
        show ?thesis by(rule subsetD [OF msteps_succeq_mono], insert sc(3) sc(4) tu, auto)
      qed
    qed
    with Cs have "(t, u) \<in> (mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  }
  then show ?thesis using mstep_def by auto
qed

lemma ERi_subset_ERw: "mstep S (E\<^sub>\<infinity> \<union> R\<^sub>\<infinity>) \<subseteq> (mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  from r_into_rtrancl have "mstep S E\<^sub>\<omega> \<subseteq> (mstep S E\<^sub>\<omega>)\<^sup>\<leftrightarrow>\<^sup>*" by auto
  with Ri_subset_ERw mstep_Un have "mstep S (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>) \<subseteq> (mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  from conversion_mono [OF this] have R: "(mstep S (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  with Ei_subset_EwRi Ri_subset_ERw show ?thesis unfolding mstep_Un [of _ "E\<^sub>\<infinity>"] by blast
qed


lemma oKBi_conversion_ERw: "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
proof
  have "rstep (E 0) \<subseteq> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
  proof
    fix s t
    assume "(s, t) \<in> rstep (E 0)"
    with ERi_subset_ERw [of "{#s, t#}"] mstep_def have st: "(s, t) \<in> (mstep {#s, t#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by force
    have "mstep {#s, t#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>) \<subseteq> rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>)" using mstep_def by force
    from conversion_mono [OF this] st show "(s, t) \<in> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  qed
  from conversion_mono [OF this] show "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by simp
next
  have "rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>) \<subseteq> (rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>*"
  proof
    fix s t
    assume "(s, t) \<in> rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>)"
    then obtain k where "(s, t) \<in> rstep (R k \<union> E k)" unfolding E\<^sub>\<omega>_def R\<^sub>\<omega>_def by blast
    with oKBi_conversion_ERi show "(s, t) \<in> (rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>*" by blast
  qed
  from conversion_mono [OF this] show "(rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>*" by simp
qed

end


locale gtotal_okb_irun = okb_irun + gtotal_reduction_order less
begin

definition Sw ("S\<^sub>\<omega>") where "S\<^sub>\<omega> \<equiv> R\<^sub>\<omega> \<union> {(s \<cdot> \<sigma>, t \<cdot> \<sigma>) | s t \<sigma>. (s, t) \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow> \<and> s \<cdot> \<sigma> \<succ> t \<cdot> \<sigma>}"

lemma gmstep_ERw_Sw: "GROUND ((mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>) \<subseteq> GROUND (((mstep S S\<^sub>\<omega>)\<^sup>\<leftrightarrow>)\<^sup>=)"
proof
  fix s t
  assume "(s, t) \<in> GROUND ((mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>)"
  then have ground: "ground s" "ground t" and mstep: "(s, t) \<in> (mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>" unfolding GROUND_def by auto
  then have dom: "\<exists>s' t'. s' \<in># S \<and> t' \<in># S \<and> s' \<succeq> s \<and> t' \<succeq> t" unfolding mstep_def by auto
  from mstep reverseTrs have mstep: "(s, t) \<in> mstep S (R\<^sub>\<omega>\<^sup>\<leftrightarrow> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)"
    unfolding mstep_def Un_iff converse_iff mem_Collect_eq split rstep_union by blast
  from mstep mstep_Un consider "(s, t) \<in> mstep S (E\<^sub>\<omega>\<^sup>\<leftrightarrow>)" | "(s, t) \<in> mstep S (R\<^sub>\<omega>\<^sup>\<leftrightarrow>)" by auto
  then show "(s, t) \<in> GROUND (((mstep S S\<^sub>\<omega>)\<^sup>\<leftrightarrow>)\<^sup>=)"
  proof (cases)
    case 1
    from ground gtotal have "s \<succ> t \<or> t \<succ> s \<or> s = t" by auto
    from 1 rstepE obtain C \<sigma> l r where C: "(l, r) \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow>" "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
      unfolding mstep_def by blast
    from ground [unfolded C ground_ctxt_apply] have "ground (l \<cdot> \<sigma>)" and "ground (r \<cdot> \<sigma>)" by auto
    from gtotal [OF this] have oriented: "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma> \<or> r \<cdot> \<sigma> \<succ> l \<cdot> \<sigma> \<or> l \<cdot> \<sigma> = r \<cdot> \<sigma>" by auto
    from C(1) have rl: "(r, l) \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow>" by auto
    have eq: "l \<cdot> \<sigma> = r \<cdot> \<sigma> \<Longrightarrow> s = t" unfolding C by auto
    have lr: "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma> \<Longrightarrow> (l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S\<^sub>\<omega>" unfolding Sw_def
      by(rule UnI2, unfold mem_Collect_eq split, insert 1 C(1), auto)
    have rl: "r \<cdot> \<sigma> \<succ> l \<cdot> \<sigma> \<Longrightarrow> (r \<cdot> \<sigma>, l \<cdot> \<sigma>) \<in> S\<^sub>\<omega>" unfolding Sw_def
      by(rule UnI2, unfold mem_Collect_eq split, insert 1 C(1), auto)
    { assume "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma> \<or> r \<cdot> \<sigma> \<succ> l \<cdot> \<sigma>"
      with lr rl have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S\<^sub>\<omega>\<^sup>\<leftrightarrow>" by auto
      with rstepI have "(s, t) \<in> (rstep S\<^sub>\<omega>)\<^sup>\<leftrightarrow>" unfolding C by auto
      with dom have "(s, t) \<in> (mstep S S\<^sub>\<omega>)\<^sup>\<leftrightarrow>" unfolding mstep_def mem_Collect_eq split by auto
      with ground have ?thesis unfolding GROUND_def by auto
    }
    with oriented eq ground show ?thesis using GROUND_def by blast
  next
    case 2
    with mstep [unfolded mstep_def] have "(s, t) \<in> (mstep S S\<^sub>\<omega>)\<^sup>\<leftrightarrow>"
      unfolding Sw_def mstep_Un mstep_def Un_iff converse_iff mem_Collect_eq split by auto
    with ground show ?thesis unfolding GROUND_def by auto
  qed
qed

lemma ground_Einf_Sw:
  assumes "(s, t) \<in> (rstep E\<^sub>\<infinity>)\<^sup>\<leftrightarrow>" and gs: "ground s" and gt: "ground t"
  shows "(s, t) \<in> (mstep {#s, t#} S\<^sub>\<omega>)\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  { fix s t
    assume a: "(s, t) \<in> rstep E\<^sub>\<infinity>" and gs: "ground s" and gt: "ground t"
    define S where "S = {#s, t#}"
    with a have "(s, t) \<in> rstep (E\<^sub>\<infinity> \<union> R\<^sub>\<infinity>)" unfolding rstep_union by auto
    then have "(s, t) \<in> mstep {#s, t#} (E\<^sub>\<infinity> \<union> R\<^sub>\<infinity>)" unfolding mstep_def by auto
    with ERi_subset_ERw have "(s, t) \<in> (mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding S_def by auto
    from this [unfolded conversion_def] have "(s, t) \<in> (mstep S S\<^sub>\<omega>)\<^sup>\<leftrightarrow>\<^sup>*"
    proof (induct)
      case base
      then show ?case by auto
    next
      case (step u v)
      note rsteps = step(2) [unfolded mstep_def Un_iff converse_iff mem_Collect_eq split]
      then have st_uv: "(s \<succeq> u \<or> t \<succeq> u) \<and> (s \<succeq> v \<or> t \<succeq> v)" unfolding S_def by force
      from ground_less have ground_leq: "\<And> s t. ground s \<Longrightarrow> s \<succeq> t \<Longrightarrow> ground t" by blast
      from st_uv ground_leq [OF gs] ground_leq [OF gt] have "ground u \<and> ground v" by metis
      with step(2) have "(u, v) \<in> GROUND ((mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>)" unfolding GROUND_def by auto
      with gmstep_ERw_Sw have uv: "(u, v) \<in> ((mstep S S\<^sub>\<omega>)\<^sup>\<leftrightarrow>)\<^sup>=" unfolding GROUND_def by auto
      with step(3) rtrancl_trans [OF step(3) [unfolded conversion_def] r_into_rtrancl] show ?case
        unfolding conversion_def by fast
    qed
    then have "(s, t) \<in> (mstep {#s, t#} S\<^sub>\<omega>)\<^sup>\<leftrightarrow>\<^sup>*" unfolding S_def by auto
  } note estep = this
  from assms(1) estep [OF _ gs gt] estep [OF _ gt gs] show ?thesis
    unfolding Un_iff add_mset_commute [of t s "{#}"] converse_iff conversion_inv by auto
qed

end

locale ordered_completion_inf =
  ordered_completion less for less :: "('a, 'b::infinite) term \<Rightarrow> ('a, 'b) term \<Rightarrow> bool" (infix "\<succ>" 50)
begin

lemma less_set_permute:
  "\<pi> \<bullet> {\<succ>} = {\<succ>}"
  apply (auto simp: )
   apply (metis inv_rule_mem_trs_simps(1) local.subst mem_Collect_eq old.prod.case term_apply_subst_Var_Rep_perm term_pt.permute_minus_cancel(1))
  by (metis case_prodI inv_rule_mem_trs_simps(1) mem_Collect_eq reduction_order.subst reduction_order_axioms term_apply_subst_Var_Rep_perm)

lemma ostep_permute:
  "(\<pi> \<bullet> s, \<pi> \<bullet> t) \<in> ostep (\<pi> \<bullet> E) (\<pi> \<bullet> R) \<longleftrightarrow> (s, t) \<in> ostep E R"
  using ordstep_permute [of \<pi> s t "{\<succ>}" "E\<^sup>\<leftrightarrow>"]
  unfolding less_set_permute
  by (simp add: eqvt [symmetric] ostep_def)

lemma oKB_permute:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "(\<pi> \<bullet> E, \<pi> \<bullet> R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (\<pi> \<bullet> E', \<pi> \<bullet> R')"
  using assms
proof (cases)
  case (deduce s t u)
  with oKB.deduce [of "\<pi> \<bullet> s" "\<pi> \<bullet> t" "\<pi> \<bullet> R" "\<pi> \<bullet> E" "\<pi> \<bullet> u"] show ?thesis
    by (simp add: eqvt) (metis (no_types, lifting) rstep_permute rstep_simps(5) rstep_union)
next
  case (orientl s t)
  with oKB.orientl [of "\<pi> \<bullet> s" "\<pi> \<bullet> t" "\<pi> \<bullet> E" "\<pi> \<bullet> R"] show ?thesis
    by (simp add: eqvt) (metis local.subst term_apply_subst_Var_Rep_perm)
next
  case (orientr t s)
  with oKB.orientr [of "\<pi> \<bullet> t" "\<pi> \<bullet> s" "\<pi> \<bullet> E" "\<pi> \<bullet> R"] show ?thesis
    by (simp add: eqvt) (metis local.subst term_apply_subst_Var_Rep_perm)
next
  case (delete s)
  with oKB.delete [of "\<pi> \<bullet> s" "\<pi> \<bullet> E" "\<pi> \<bullet> R"] show ?thesis by (simp add: eqvt)
next
  case (compose t u s)
  with oKB.compose [of "\<pi> \<bullet> t" "\<pi> \<bullet> u" "\<pi> \<bullet> E" "\<pi> \<bullet> R" "\<pi> \<bullet> s"] show ?thesis
    using ostep_permute [of \<pi> t u E "R - {(s, t)}"] by (simp add: eqvt)
next
  case (simplifyl s u t)
  with oKB.simplifyl [of "\<pi> \<bullet> s" "\<pi> \<bullet> u" "\<pi> \<bullet> E" "\<pi> \<bullet> t" "\<pi> \<bullet> R"] show ?thesis
    using ostep_permute [of \<pi> s u "E - {(s, t)}" R] by (simp add: eqvt)
next
  case (simplifyr t u s)
  with oKB.simplifyr [of "\<pi> \<bullet> t" "\<pi> \<bullet> u" "\<pi> \<bullet> E" "\<pi> \<bullet> s" "\<pi> \<bullet> R"] show ?thesis
    using ostep_permute [of \<pi> t u "E - {(s, t)}" R] by (simp add: eqvt)
next
  case (collapse t u s)
  with oKB.collapse [of "\<pi> \<bullet> t" "\<pi> \<bullet> u" "\<pi> \<bullet> E" "\<pi> \<bullet> R" "\<pi> \<bullet> s"] show ?thesis
    using ostep_permute [of \<pi> t u "E" "R - {(t, s)}"] by (simp add: eqvt)
qed

inductive
  oKB' :: "('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> ('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> bool"
  where
    deduce: "(s, t) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>) \<Longrightarrow> (s, u) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>) \<Longrightarrow> oKB' (E, R) (E \<union> {(t, u)}, R)" |
    orientl: "s \<succ> t \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> oKB' (E, R) (E - {(s, t)}, R \<union> {(p \<bullet> s, p \<bullet> t)})" |
    orientr: "t \<succ> s \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> oKB' (E, R) (E - {(s, t)}, R \<union> {(p \<bullet> t, p \<bullet> s)})" |
    delete: "(s, s) \<in> E \<Longrightarrow> oKB' (E, R) (E - {(s, s)}, R)" |
    compose: "(t, u) \<in> ostep E (R - {(s, t)}) \<Longrightarrow> (s, t) \<in> R \<Longrightarrow> oKB' (E, R) (E, (R - {(s, t)}) \<union> {(p \<bullet> s, p \<bullet> u)})" |
    simplifyl: "(s, u) \<in> ostep (E - {(s, t)}) R \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> oKB' (E, R) ((E - {(s, t)}) \<union> {(p \<bullet> u, p \<bullet> t)}, R)" |
    simplifyr: "(t, u) \<in> ostep (E - {(s, t)}) R \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> oKB' (E, R) ((E - {(s, t)}) \<union> {(p \<bullet> s, p \<bullet> u)}, R)" |
    collapse: "(t, u) \<in> ostep E (R - {(t, s)}) \<Longrightarrow> (t, s) \<in> R \<Longrightarrow> oKB' (E, R) (E \<union> {(p \<bullet> u, p \<bullet> s)}, R - {(t, s)})"

end

definition "E_ord ord E = {(s \<cdot> \<sigma>, t \<cdot> \<sigma>) | s t \<sigma>. (s, t) \<in> E\<^sup>\<leftrightarrow> \<and> ord (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)}"

locale gtotal_okb_irun_inf =
  ordered_completion_inf less +
  gtotal_okb_irun less for less :: "('a, 'b::infinite) term \<Rightarrow> ('a, 'b) term \<Rightarrow> bool" (infix "\<succ>" 50)
begin

sublocale gtotal_reduction_order_inf ..

definition "PCP_ext RR = {(u, v) | r r' p \<mu> u v.
    ooverlap {\<succ>} RR r r' p \<mu> u v \<and>
    (\<forall>u \<lhd> fst r \<cdot> \<mu>. u \<in> NF (ordstep {\<succ>} RR))}"

context
  fixes \<R> \<E> :: "('a, 'b) trs"
  assumes rules: "\<R> \<subseteq> {\<succ>}"
begin

abbreviation "\<S> \<equiv> \<R> \<union> E_ord (op \<succ>) \<E>"

lemma perm_R_less: "\<pi> \<bullet> (l, r) \<in> \<R> \<Longrightarrow> l \<succ> r"
  using rules by (auto simp: eqvt perm_less)

lemma rstep_S_ordstep:
  assumes "(s, t) \<in> rstep \<S>"
  shows "(s, t) \<in> ordstep {\<succ>} \<S>"
  using assms
proof (cases rule: rstepE)
  case (rstep C \<sigma> l r)
  then show ?thesis
    using rules
    by (intro ordstep.intros [where C = C and l = l and r = r and \<sigma> = \<sigma>])
      (auto dest: subst simp: E_ord_def)
qed

lemma ordstep_subset_S:
  "ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>) \<subseteq> ordstep {\<succ>} \<S>"
proof -
  { fix s t assume "(s, t) \<in> ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)"
    then obtain C l r \<sigma> where "(l, r) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>"
      and [simp]: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>" "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma>" by (cases) auto
    then consider "(l, r) \<in> \<R>" | "(l, r) \<in> \<E>\<^sup>\<leftrightarrow>" by blast
    then have "(s, t) \<in> ordstep {\<succ>} \<S>"
    proof (cases)
      case 1
      then show ?thesis by (auto intro: ordstep.intros)
    next
      case 2
      then show ?thesis
        by (intro ordstep.intros [where C = C and \<sigma> = Var and l = "l \<cdot> \<sigma>" and r = "r \<cdot> \<sigma>"])
          (force simp: E_ord_def)+
    qed }
  then show ?thesis by auto
qed

lemma NF_ordstep_S_subset: "NF (ordstep {\<succ>} \<S>) \<subseteq> NF (ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))"
  using ordstep_subset_S by blast

lemma ordstep_S_conv [simp]:
  "ordstep {\<succ>} \<S> = rstep \<S>"
  using rstep_S_ordstep by (auto elim: ordstep.cases)

lemma perm_S_ordstep:
  assumes "\<pi> \<bullet> (s, t) \<in> \<S>"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ordstep {\<succ>} \<S>"
  using assms
  by (intro ordstep.intros [where C = \<box> and \<sigma> = "sop (-\<pi>) \<circ>\<^sub>s \<sigma>" and l = "\<pi> \<bullet> s" and r = "\<pi> \<bullet> t"])
    (auto simp: eqvt perm_R_less E_ord_def subst, (metis perm_less subst)+)

lemma perm_R_ordstep:
  assumes "\<pi> \<bullet> (s, t) \<in> \<R>"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)"
  using assms
  by (intro ordstep.intros [where C = \<box> and \<sigma> = "sop (-\<pi>) \<circ>\<^sub>s \<sigma>" and l = "\<pi> \<bullet> s" and r = "\<pi> \<bullet> t"])
    (auto simp: eqvt perm_R_less subst)

lemma E_ordstep:
  assumes "(s, t) \<in> \<E>\<^sup>\<leftrightarrow>" and "s \<cdot> \<sigma> \<succ> t \<cdot> \<sigma>"
  shows "(s \<cdot> \<sigma> \<circ>\<^sub>s \<tau>, t \<cdot> \<sigma> \<circ>\<^sub>s \<tau>) \<in> ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)"
  using assms
  by (intro ordstep.intros [where C = \<box> and \<sigma> = "\<sigma> \<circ>\<^sub>s \<tau>" and l = "s" and r = "t"])
    (auto simp: subst)

lemma PCP_xPCP:
  assumes "ground s" and "ground t"
    and "(s, t) \<in> (rstep (PCP \<S>))\<^sup>\<leftrightarrow>"
  shows "(s, t) \<in> (rstep (PCP_ext (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)))\<^sup>\<leftrightarrow> \<union> (rstep \<S>)\<^sup>\<down>"
    (is "_ \<in> ?R")
proof -
  have S: "\<And>s t. (s, t) \<in> \<S> \<Longrightarrow> s \<succ> t \<or> (t, s) \<in> \<S>"
    using rules by (auto simp: E_ord_def)
  have RE: "\<And>s t. (s, t) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow> \<Longrightarrow> s \<succ> t \<or> (t, s) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>"
    using rules by (auto)
  let ?S = "ordstep {\<succ>} \<S>"
  let ?Eo = "E_ord (op \<succ>) \<E>"
  let ?RE = "ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)"
  { fix s t and \<sigma> :: "('a, 'b) subst"
    assume "(s, t) \<in> PCP \<S>" and ground: "ground (s \<cdot> \<sigma>)" "ground (t \<cdot> \<sigma>)"
    then obtain \<mu> l\<^sub>1 r\<^sub>1 p l\<^sub>2 r\<^sub>2 where o: "overlap \<S> \<S> (l\<^sub>1, r\<^sub>1) p (l\<^sub>2, r\<^sub>2)"
      and "the_mgu l\<^sub>1 (l\<^sub>2 |_ p) = \<mu>"
      and s: "s = replace_at l\<^sub>2 p r\<^sub>1 \<cdot> \<mu>"
      and t: "t = r\<^sub>2 \<cdot> \<mu>"
      and NF: "\<forall>u \<lhd> l\<^sub>1 \<cdot> \<mu>. u \<in> NF (rstep \<S>)"
      by (auto simp: PCP_def the_mgu_def split: option.splits)
    then have mgu: "mgu l\<^sub>1 (l\<^sub>2 |_ p) = Some \<mu>"
      by (auto simp: the_mgu_def unifiers_def overlap_def simp del: mgu.simps split: option.splits dest!: mgu_complete)
    from mgu_sound [OF this]
    have "l\<^sub>1 \<cdot> \<mu> = (l\<^sub>2 |_ p) \<cdot> \<mu>" by (auto simp: is_imgu_def)
    moreover
    define C and \<tau> where "C = ctxt_of_pos_term p (l\<^sub>2 \<cdot> \<mu>) \<cdot>\<^sub>c \<sigma>" and "\<tau> = \<mu> \<circ>\<^sub>s \<sigma>"
    ultimately
    have eq: "\<box>\<langle>l\<^sub>2 \<cdot> \<tau>\<rangle> = C\<langle>l\<^sub>1 \<cdot> \<tau>\<rangle>"
      using o
      by simp (metis ctxt_of_pos_term_subst ctxt_supt_id fst_conv overlap_source_eq(1) subst_apply_term_ctxt_apply_distrib)

    have s': "s \<cdot> \<sigma> = C\<langle>r\<^sub>1 \<cdot> \<tau>\<rangle>"
      using o by (simp add: s C_def \<tau>_def overlap_def ctxt_of_pos_term_subst funposs_imp_poss)

    have t': "t \<cdot> \<sigma> = \<box>\<langle>r\<^sub>2 \<cdot> \<tau>\<rangle>" by (simp add: t \<tau>_def)

    have p: "p \<in> funposs l\<^sub>2" using o by (auto simp: overlap_def)
    then have [simp]: "hole_pos C = p" by (auto simp: C_def funposs_imp_poss)

    have disj: "vars_rule (l\<^sub>2, r\<^sub>2) \<inter> vars_rule (l\<^sub>1, r\<^sub>1) = {}" using o by (auto simp: overlap_def)

    { fix \<pi>\<^sub>1 and \<pi>\<^sub>2 assume *: "\<pi>\<^sub>1 \<bullet> (l\<^sub>1, r\<^sub>1) \<in> \<R>" "\<pi>\<^sub>2 \<bullet> (l\<^sub>2, r\<^sub>2) \<in> \<R>"
      then have rule1: "\<pi>\<^sub>1 \<bullet> (l\<^sub>1, r\<^sub>1) \<in> \<S>" and rule2: "\<pi>\<^sub>2 \<bullet> (l\<^sub>2, r\<^sub>2) \<in> \<S>" by auto
      have "(s \<cdot> \<sigma> , t \<cdot> \<sigma>) \<in> ?R"
      proof (cases "hole_pos C \<in> funposs l\<^sub>2")
        case False
        from non_ooverlap_GROUND_joinable [OF S eq refl False _ _ _ s' t' ground rule2]
          and perm_S_ordstep [OF rule1, of \<tau>] and perm_S_ordstep [OF rule2, of \<tau>]
        show ?thesis by (auto) (metis GROUND_subset Un_absorb1 Un_iff join_mono)
      next
        case True
        then have "ooverlap {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>) (l\<^sub>1, r\<^sub>1) (l\<^sub>2, r\<^sub>2) p \<mu> s t"
          using * and disj and p [THEN funposs_imp_poss] and irrefl
          by (auto simp: ooverlap_def mgu s t ctxt_of_pos_term_subst simp del: mgu.simps)
            (auto dest!: perm_R_less dest: subst trans)
        moreover have "\<forall>u \<lhd> l\<^sub>1 \<cdot> \<mu>. u \<in> NF (ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))"
          using NF and NF_ordstep_S_subset by auto
        ultimately have "(s, t) \<in> PCP_ext (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)"
          by  (force simp: PCP_ext_def)
        then show ?thesis by blast
      qed }
    moreover
    { fix \<pi>\<^sub>1 and \<pi>\<^sub>2 assume rule1: "\<pi>\<^sub>1 \<bullet> (l\<^sub>1, r\<^sub>1) \<in> \<R>" and "\<pi>\<^sub>2 \<bullet> (l\<^sub>2, r\<^sub>2) \<in> ?Eo"
      then obtain u\<^sub>2 v\<^sub>2 \<tau>\<^sub>2' where rule2: "(u\<^sub>2, v\<^sub>2) \<in> \<E>\<^sup>\<leftrightarrow>"
        and "\<pi>\<^sub>2 \<bullet> l\<^sub>2 = u\<^sub>2 \<cdot> \<tau>\<^sub>2'" and "\<pi>\<^sub>2 \<bullet> r\<^sub>2 = v\<^sub>2 \<cdot> \<tau>\<^sub>2'"
        and "\<pi>\<^sub>2 \<bullet> l\<^sub>2 \<succ> \<pi>\<^sub>2 \<bullet> r\<^sub>2" and "u\<^sub>2 \<cdot> \<tau>\<^sub>2' \<succ> v\<^sub>2 \<cdot> \<tau>\<^sub>2'"
        by (auto simp: E_ord_def eqvt)
      moreover define \<tau>\<^sub>2 where "\<tau>\<^sub>2 = \<tau>\<^sub>2' \<circ>\<^sub>s sop (-\<pi>\<^sub>2)"
      ultimately have [simp]: "l\<^sub>2 = u\<^sub>2 \<cdot> \<tau>\<^sub>2" "r\<^sub>2 = v\<^sub>2 \<cdot> \<tau>\<^sub>2"
        and gr2: "u\<^sub>2 \<cdot> \<tau>\<^sub>2 \<succ> v\<^sub>2 \<cdot> \<tau>\<^sub>2" by (auto simp: term_pt.permute_flip perm_less)
      from eq have eq': "\<box>\<langle>u\<^sub>2 \<cdot> \<tau>\<^sub>2 \<circ>\<^sub>s \<tau>\<rangle> = C\<langle>l\<^sub>1 \<cdot> \<tau>\<rangle>" by simp

      have rule2': "0 \<bullet> (u\<^sub>2, v\<^sub>2) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>" using \<open>(u\<^sub>2, v\<^sub>2) \<in> \<E>\<^sup>\<leftrightarrow>\<close> by (auto)
      have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ?R"
      proof (cases "hole_pos C \<in> funposs u\<^sub>2")
        case False

        from t' have t'': "t \<cdot> \<sigma> = \<box>\<langle>v\<^sub>2 \<cdot> \<tau>\<^sub>2 \<circ>\<^sub>s \<tau>\<rangle>" by simp
        from non_ooverlap_GROUND_joinable [OF RE eq' refl False _ _ _ s' t'' ground rule2']
          and perm_R_ordstep [OF rule1, of \<tau>] and E_ordstep [OF rule2 gr2, of \<tau>]
        have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (GROUND (ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)))\<^sup>\<down>" by auto
        then show ?thesis
          using join_mono [OF ordstep_subset_S] and GROUND_subset
          by (metis (no_types, lifting) UnI2 join_mono ordstep_S_conv set_mp)
      next
        case True
        obtain \<pi> where disj: "vars_rule (l\<^sub>1, r\<^sub>1) \<inter> vars_rule (\<pi> \<bullet> u\<^sub>2, \<pi> \<bullet> v\<^sub>2) = {}"
          using vars_rule_disjoint unfolding eqvt [symmetric] by blast
        have rule2'': "-\<pi> \<bullet> (\<pi> \<bullet> u\<^sub>2, \<pi> \<bullet> v\<^sub>2) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>"
          using rule2' by (simp)
        then have rule2'': "\<exists>p. p \<bullet> (\<pi> \<bullet> u\<^sub>2, \<pi> \<bullet> v\<^sub>2) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>" ..
        define \<nu> where "\<nu> x = (if x \<in> vars_rule (l\<^sub>1, r\<^sub>1) then Var x else (sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>2) x)" for x
        have 1: "\<pi> \<bullet> u\<^sub>2 \<cdot> \<nu> = u\<^sub>2 \<cdot> \<tau>\<^sub>2"
        proof -
          have "\<pi> \<bullet> u\<^sub>2 \<cdot> \<nu> = \<pi> \<bullet> u\<^sub>2 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>2"
            using disj by (unfold term_subst_eq_conv) (auto simp add: \<nu>_def vars_defs)
          then show ?thesis by simp
        qed
        have 2: "\<pi> \<bullet> v\<^sub>2 \<cdot> \<nu> = v\<^sub>2 \<cdot> \<tau>\<^sub>2"
        proof -
          have "\<pi> \<bullet> v\<^sub>2 \<cdot> \<nu> = \<pi> \<bullet> v\<^sub>2 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>2"
            using disj by (unfold term_subst_eq_conv) (auto simp add: \<nu>_def vars_defs)
          then show ?thesis by simp
        qed
        have [simp]: "r\<^sub>1 \<cdot> \<nu> = r\<^sub>1"
        proof -
          have "r\<^sub>1 \<cdot> \<nu> = r\<^sub>1 \<cdot> Var" by (unfold term_subst_eq_conv) (simp add: \<nu>_def vars_defs)
          then show ?thesis by simp
        qed
        have r\<^sub>2: "r\<^sub>2 = \<pi> \<bullet> v\<^sub>2 \<cdot> \<nu>"
        proof -
          have "\<pi> \<bullet> v\<^sub>2 \<cdot> \<nu> = \<pi> \<bullet> v\<^sub>2 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>2"
            using disj
            by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def eqvt)
          then show ?thesis using p and True
            by (auto simp: eqvt dest!: funposs_imp_poss)
        qed
        have l\<^sub>2: "l\<^sub>2 = \<pi> \<bullet> u\<^sub>2 \<cdot> \<nu>"
        proof -
          have "\<pi> \<bullet> u\<^sub>2 \<cdot> \<nu> = \<pi> \<bullet> u\<^sub>2 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>2"
            using disj
            by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def eqvt)
          then show ?thesis using p and True
            by (auto simp: eqvt dest!: funposs_imp_poss)
        qed
        then have "l\<^sub>2 |_ p = (\<pi> \<bullet> u\<^sub>2 |_ p) \<cdot> \<nu>"
          using True by (auto dest!: funposs_imp_poss)
        moreover
        have "l\<^sub>1 \<cdot> \<nu> \<circ>\<^sub>s \<mu> = l\<^sub>1 \<cdot> \<mu>"
          by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def)
        ultimately
        have *: "l\<^sub>1 \<cdot> \<nu> \<circ>\<^sub>s \<mu> = (\<pi> \<bullet> u\<^sub>2 |_ p) \<cdot> \<nu> \<circ>\<^sub>s \<mu>" using \<open>l\<^sub>1 \<cdot> \<mu> = (l\<^sub>2 |_ p) \<cdot> \<mu>\<close> by simp
        then obtain \<mu>' where mgu: "mgu l\<^sub>1 (\<pi> \<bullet> u\<^sub>2 |_ p) = Some \<mu>'"
          using mgu_complete by (auto simp: unifiers_def simp del: subst_subst_compose)
        from mgu_sound [OF this, THEN is_imgu_imp_is_mgu] and * obtain \<delta> where \<mu>': " \<nu> \<circ>\<^sub>s \<mu> = \<mu>' \<circ>\<^sub>s \<delta>"
          by (auto simp: is_mgu_def unifiers_def simp del: subst_subst_compose)
        have l\<^sub>1: "l\<^sub>1 \<cdot> \<mu> = l\<^sub>1 \<cdot> \<mu>' \<circ>\<^sub>s \<delta>"
          by (unfold term_subst_eq_conv, fold \<mu>') (simp add: subst_compose \<nu>_def vars_defs)
        let ?s = "(ctxt_of_pos_term p (\<pi> \<bullet> u\<^sub>2 \<cdot> \<mu>'))\<langle>r\<^sub>1 \<cdot> \<mu>'\<rangle>"
        let ?t = "\<pi> \<bullet> v\<^sub>2 \<cdot> \<mu>'"
        have 3: "t \<cdot> \<mu>' \<cdot> \<delta> = t \<cdot> \<nu> \<cdot> \<mu>" for t by (fold subst_subst_compose) (simp add: \<mu>')
        have "p \<in> poss (\<pi> \<bullet> u\<^sub>2)" using True by (simp add: eqvt funposs_imp_poss)
        then have 4: "s \<cdot> \<sigma> = ?s \<cdot> \<delta> \<cdot> \<sigma>"
          by (simp add: s ctxt_of_pos_term_subst [symmetric] 1 3)
        have 5: "t \<cdot> \<sigma> = ?t \<cdot> \<delta> \<cdot> \<sigma>" by (simp add: t 2 3)

        have "\<not> \<pi> \<bullet> v\<^sub>2 \<cdot> \<mu>' \<succ> \<pi> \<bullet> u\<^sub>2 \<cdot> \<mu>'" (is "\<not> ?v \<cdot> _ \<succ> ?u \<cdot> _")
        proof
          presume "\<not> ?thesis"
          then have "?v \<cdot> \<nu> \<circ>\<^sub>s \<mu> \<succ> ?u \<cdot> \<nu> \<circ>\<^sub>s \<mu>"
            unfolding \<mu>' by (simp add: subst)
          then have "r\<^sub>2 \<cdot> \<mu> \<succ> l\<^sub>2 \<cdot> \<mu>" unfolding l\<^sub>2 r\<^sub>2 by simp
          moreover have "l\<^sub>2 \<succ> r\<^sub>2" using \<open>\<pi>\<^sub>2 \<bullet> l\<^sub>2 \<succ> \<pi>\<^sub>2 \<bullet> r\<^sub>2\<close> by (simp add: perm_less)
          ultimately show False using irrefl by (auto dest: subst trans)
        qed simp
        then have "ooverlap {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>) (l\<^sub>1, r\<^sub>1) (\<pi> \<bullet> u\<^sub>2, \<pi> \<bullet> v\<^sub>2) p \<mu>' ?s ?t"
          using mgu and disj and rule1 and rule2'' and True and irrefl
          by (auto simp: ooverlap_def perm_less simp del: mgu.simps)
            (auto dest!: perm_R_less dest: subst trans simp del: mgu.simps)
        moreover have "\<forall>u \<lhd> l\<^sub>1 \<cdot> \<mu>'. u \<in> NF (ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))"
          using NF and NF_ordstep_S_subset
          unfolding l\<^sub>1
          by auto (meson NF_instance contra_subsetD instance_no_supt_imp_no_supt)
        ultimately have "(?s, ?t) \<in> PCP_ext (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)" by (force simp: PCP_ext_def)
        then show ?thesis
          unfolding 4 and 5 by blast
      qed }
    moreover
    { fix \<pi>\<^sub>1 and \<pi>\<^sub>2 assume *: "\<pi>\<^sub>1 \<bullet> (l\<^sub>1, r\<^sub>1) \<in> ?Eo" and rule2: "\<pi>\<^sub>2 \<bullet> (l\<^sub>2, r\<^sub>2) \<in> \<R>"
      then have rule1: "\<pi>\<^sub>1 \<bullet> (l\<^sub>1, r\<^sub>1) \<in> \<S>" and rule2': "\<pi>\<^sub>2 \<bullet> (l\<^sub>2, r\<^sub>2) \<in> \<S>" by auto
      have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ?R"
      proof (cases "hole_pos C \<in> funposs l\<^sub>2")
        case False
        from non_ooverlap_GROUND_joinable [OF S eq refl False _ _ _ s' t' ground rule2']
          and perm_S_ordstep [OF rule1, of \<tau>] and perm_S_ordstep [OF rule2', of \<tau>]
        show ?thesis by (auto) (metis GROUND_subset Un_absorb1 Un_iff join_mono)
      next
        from * obtain u\<^sub>1 v\<^sub>1 \<tau>\<^sub>1' where rule1: "(u\<^sub>1, v\<^sub>1) \<in> \<E>\<^sup>\<leftrightarrow>"
          and "\<pi>\<^sub>1 \<bullet> l\<^sub>1 = u\<^sub>1 \<cdot> \<tau>\<^sub>1'" and "\<pi>\<^sub>1 \<bullet> r\<^sub>1 = v\<^sub>1 \<cdot> \<tau>\<^sub>1'"
          and "\<pi>\<^sub>1 \<bullet> l\<^sub>1 \<succ> \<pi>\<^sub>1 \<bullet> r\<^sub>1" and "u\<^sub>1 \<cdot> \<tau>\<^sub>1' \<succ> v\<^sub>1 \<cdot> \<tau>\<^sub>1'"
          by (auto simp: E_ord_def eqvt)
        moreover define \<tau>\<^sub>1 where "\<tau>\<^sub>1 = \<tau>\<^sub>1' \<circ>\<^sub>s sop (-\<pi>\<^sub>1)"
        ultimately have [simp]: "l\<^sub>1 = u\<^sub>1 \<cdot> \<tau>\<^sub>1" "r\<^sub>1 = v\<^sub>1 \<cdot> \<tau>\<^sub>1"
          and gr2: "u\<^sub>1 \<cdot> \<tau>\<^sub>1 \<succ> v\<^sub>1 \<cdot> \<tau>\<^sub>1" by (auto simp: term_pt.permute_flip perm_less)
        case True
        obtain \<pi> where disj: "vars_rule (\<pi> \<bullet> u\<^sub>1, \<pi> \<bullet> v\<^sub>1) \<inter> vars_rule (l\<^sub>2, r\<^sub>2) = {}"
          using vars_rule_disjoint unfolding eqvt [symmetric] by blast
        have rule1'': "-\<pi> \<bullet> (\<pi> \<bullet> u\<^sub>1, \<pi> \<bullet> v\<^sub>1) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>"
          using rule1 by (simp)
        then have rule1'': "\<exists>p. p \<bullet> (\<pi> \<bullet> u\<^sub>1, \<pi> \<bullet> v\<^sub>1) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>" ..
        define \<nu> where "\<nu> x = (if x \<in> vars_rule (l\<^sub>2, r\<^sub>2) then Var x else (sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>1) x)" for x
        have [simp]: "l\<^sub>2 \<cdot> \<nu> = l\<^sub>2 \<cdot> Var" by (unfold term_subst_eq_conv) (simp add: \<nu>_def vars_defs)
        then have [simp]: "(l\<^sub>2 |_ p) \<cdot> \<nu> = l\<^sub>2 |_ p"
          using True by (auto dest!: funposs_imp_poss) (metis \<open>l\<^sub>2 \<cdot> \<nu> = l\<^sub>2 \<cdot> Var\<close> subst.cop_nil subt_at_subst)

        have [simp]: "r\<^sub>2 \<cdot> \<nu> = r\<^sub>2 \<cdot> Var" by (unfold term_subst_eq_conv) (simp add: \<nu>_def vars_defs)
        have [simp]: "\<pi> \<bullet> u\<^sub>1 \<cdot> \<nu> = u\<^sub>1 \<cdot> \<tau>\<^sub>1"
        proof -
          have "\<pi> \<bullet> u\<^sub>1 \<cdot> \<nu> = \<pi> \<bullet> u\<^sub>1 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>1"
            using disj
            by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def eqvt)
          then show ?thesis by simp
        qed
        have [simp]: "\<pi> \<bullet> v\<^sub>1 \<cdot> \<nu> = v\<^sub>1 \<cdot> \<tau>\<^sub>1"
        proof -
          have "\<pi> \<bullet> v\<^sub>1 \<cdot> \<nu> = \<pi> \<bullet> v\<^sub>1 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>1"
            using disj
            by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def eqvt)
          then show ?thesis by simp
        qed
        have l\<^sub>1: "l\<^sub>1 = \<pi> \<bullet> u\<^sub>1 \<cdot> \<nu>"
        proof -
          have "\<pi> \<bullet> u\<^sub>1 \<cdot> \<nu> = \<pi> \<bullet> u\<^sub>1 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>1"
            using disj
            by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def eqvt)
          then show ?thesis using p and True
            by (auto simp: eqvt dest!: funposs_imp_poss)
        qed
        have r\<^sub>1: "r\<^sub>1 = \<pi> \<bullet> v\<^sub>1 \<cdot> \<nu>"
        proof -
          have "\<pi> \<bullet> v\<^sub>1 \<cdot> \<nu> = \<pi> \<bullet> v\<^sub>1 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>1"
            using disj
            by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def eqvt)
          then show ?thesis using p and True
            by (auto simp: eqvt dest!: funposs_imp_poss)
        qed
        have *: "\<pi> \<bullet> u\<^sub>1 \<cdot> \<nu> \<circ>\<^sub>s \<mu> = (l\<^sub>2 |_ p) \<cdot> \<nu> \<circ>\<^sub>s \<mu>" using \<open>l\<^sub>1 \<cdot> \<mu> = (l\<^sub>2 |_ p) \<cdot> \<mu>\<close> by (simp)
        then obtain \<mu>' where mgu: "mgu (\<pi> \<bullet> u\<^sub>1) (l\<^sub>2 |_ p) = Some \<mu>'"
          using mgu_complete by (auto simp: unifiers_def simp del: subst_subst_compose)
        from mgu_sound [OF this, THEN is_imgu_imp_is_mgu] and * obtain \<delta> where \<mu>': " \<nu> \<circ>\<^sub>s \<mu> = \<mu>' \<circ>\<^sub>s \<delta>"
          by (auto simp: is_mgu_def unifiers_def simp del: subst_subst_compose)

        have 1: "t \<cdot> \<mu>' \<cdot> \<delta> = t \<cdot> \<nu> \<cdot> \<mu>" for t
          by (fold subst_subst_compose) (simp add: \<mu>')
        let ?s = "(ctxt_of_pos_term p (l\<^sub>2 \<cdot> \<mu>'))\<langle>\<pi> \<bullet> v\<^sub>1 \<cdot> \<mu>'\<rangle>"
        let ?t = "r\<^sub>2 \<cdot> \<mu>'"
        have 2: "s \<cdot> \<sigma> = ?s \<cdot> \<delta> \<cdot> \<sigma>"
          using True [THEN funposs_imp_poss]
          by (simp add: s ctxt_of_pos_term_subst [symmetric] 1)
        have 3: "t \<cdot> \<sigma> = ?t \<cdot> \<delta> \<cdot> \<sigma>" by (simp add: t 1)

        have "\<not> \<pi> \<bullet> v\<^sub>1 \<cdot> \<mu>' \<succ> \<pi> \<bullet> u\<^sub>1 \<cdot> \<mu>'"
          (is "\<not> ?v \<cdot> _ \<succ> ?u \<cdot> _")
        proof
          presume "\<not> ?thesis"
          then have "?v \<cdot> \<nu> \<circ>\<^sub>s \<mu> \<succ> ?u \<cdot> \<nu> \<circ>\<^sub>s \<mu>"
            unfolding \<mu>' by (simp add: subst)
          then have "r\<^sub>1 \<cdot> \<mu> \<succ> l\<^sub>1 \<cdot> \<mu>" unfolding l\<^sub>1 r\<^sub>1 by simp
          moreover have "l\<^sub>1 \<succ> r\<^sub>1" using \<open>\<pi>\<^sub>1 \<bullet> l\<^sub>1 \<succ> \<pi>\<^sub>1 \<bullet> r\<^sub>1\<close> by (simp add: perm_less)
          ultimately show False using irrefl by (auto dest: subst trans)
        qed simp
        then have "ooverlap {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>) (\<pi> \<bullet> u\<^sub>1, \<pi> \<bullet> v\<^sub>1) (l\<^sub>2, r\<^sub>2) p \<mu>' ?s ?t"
          using mgu and disj and rule1'' and rule2 and True and irrefl
          by (auto simp: ooverlap_def perm_less) (auto dest!: perm_R_less dest: subst trans)
        moreover have "\<forall>u \<lhd> \<pi> \<bullet> u\<^sub>1 \<cdot> \<mu>'. u \<in> NF (ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))"
          using NF and NF_ordstep_S_subset
          unfolding l\<^sub>1 and 1 [symmetric]
          by auto (meson NF_instance contra_subsetD instance_no_supt_imp_no_supt)
        ultimately have "(?s, ?t) \<in> PCP_ext (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)" by (force simp: PCP_ext_def)
        then show ?thesis
          unfolding 2 and 3 by blast
      qed }
    moreover
    { fix \<pi>\<^sub>1 and \<pi>\<^sub>2 assume "\<pi>\<^sub>1 \<bullet> (l\<^sub>1, r\<^sub>1) \<in> ?Eo" and "\<pi>\<^sub>2 \<bullet> (l\<^sub>2, r\<^sub>2) \<in> ?Eo"
      then obtain u\<^sub>1 v\<^sub>1 \<tau>\<^sub>1' and u\<^sub>2 v\<^sub>2 \<tau>\<^sub>2'
        where rule1: "(u\<^sub>1, v\<^sub>1) \<in> \<E>\<^sup>\<leftrightarrow>" and "\<pi>\<^sub>1 \<bullet> l\<^sub>1 = u\<^sub>1 \<cdot> \<tau>\<^sub>1'" and "\<pi>\<^sub>1 \<bullet> r\<^sub>1 = v\<^sub>1 \<cdot> \<tau>\<^sub>1'"
          and "\<pi>\<^sub>1 \<bullet> l\<^sub>1 \<succ> \<pi>\<^sub>1 \<bullet> r\<^sub>1"
          and rule2: "(u\<^sub>2, v\<^sub>2) \<in> \<E>\<^sup>\<leftrightarrow>" and "\<pi>\<^sub>2 \<bullet> l\<^sub>2 = u\<^sub>2 \<cdot> \<tau>\<^sub>2'" and "\<pi>\<^sub>2 \<bullet> r\<^sub>2 = v\<^sub>2 \<cdot> \<tau>\<^sub>2'"
          and "\<pi>\<^sub>2 \<bullet> l\<^sub>2 \<succ> \<pi>\<^sub>2 \<bullet> r\<^sub>2"
        by (force simp: E_ord_def eqvt)
      moreover
      define \<tau>\<^sub>1 and \<tau>\<^sub>2 where "\<tau>\<^sub>1 = \<tau>\<^sub>1' \<circ>\<^sub>s sop (-\<pi>\<^sub>1)" and "\<tau>\<^sub>2 = \<tau>\<^sub>2' \<circ>\<^sub>s sop (-\<pi>\<^sub>2)"
      ultimately
      have [simp]: "u\<^sub>1 \<cdot> \<tau>\<^sub>1 = l\<^sub>1" "v\<^sub>1 \<cdot> \<tau>\<^sub>1 = r\<^sub>1" "u\<^sub>2 \<cdot> \<tau>\<^sub>2 = l\<^sub>2" "v\<^sub>2 \<cdot> \<tau>\<^sub>2 = r\<^sub>2"
        and gr1: "u\<^sub>1 \<cdot> \<tau>\<^sub>1 \<succ> v\<^sub>1 \<cdot> \<tau>\<^sub>1" and gr2: "u\<^sub>2 \<cdot> \<tau>\<^sub>2 \<succ> v\<^sub>2 \<cdot> \<tau>\<^sub>2"
        by (auto simp: term_pt.permute_flip [symmetric] perm_less)

      have rule2': "0 \<bullet> (u\<^sub>2, v\<^sub>2) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>" using rule2 by simp

      have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ?R"
      proof (cases "hole_pos C \<in> funposs u\<^sub>2")
        case False

        from eq have eq': "\<box>\<langle>u\<^sub>2 \<cdot> \<tau>\<^sub>2 \<circ>\<^sub>s \<tau>\<rangle> = C\<langle>u\<^sub>1 \<cdot> \<tau>\<^sub>1 \<circ>\<^sub>s \<tau>\<rangle>" by simp
        from s' have s'': "s \<cdot> \<sigma> = C\<langle>v\<^sub>1 \<cdot> \<tau>\<^sub>1 \<circ>\<^sub>s \<tau>\<rangle>" by simp
        from t' have t'': "t \<cdot> \<sigma> = \<box>\<langle>v\<^sub>2 \<cdot> \<tau>\<^sub>2 \<circ>\<^sub>s \<tau>\<rangle>" by simp
        from non_ooverlap_GROUND_joinable [OF RE eq' refl False _ _ _ s'' t'' ground rule2']
          and E_ordstep [OF rule1 gr1, of \<tau>] and E_ordstep [OF rule2 gr2, of \<tau>]
        have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (GROUND (ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)))\<^sup>\<down>" by auto
        then show ?thesis
          using join_mono [OF ordstep_subset_S] and GROUND_subset
          by (metis (no_types, lifting) UnI2 join_mono ordstep_S_conv set_mp)
      next
        case True
        obtain \<pi> where disj: "vars_rule (\<pi> \<bullet> u\<^sub>1, \<pi> \<bullet> v\<^sub>1) \<inter> vars_rule (u\<^sub>2, v\<^sub>2) = {}"
          using vars_rule_disjoint unfolding eqvt [symmetric] by blast
        have "-\<pi> \<bullet> (\<pi> \<bullet> u\<^sub>1, \<pi> \<bullet> v\<^sub>1) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>" using rule1 by simp
        then have rule1': "\<exists>p. p \<bullet> (\<pi> \<bullet> u\<^sub>1, \<pi> \<bullet> v\<^sub>1) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>" ..
        have rule2'': "\<exists>p. p \<bullet> (u\<^sub>2, v\<^sub>2) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>" using rule2' ..

        define \<nu> where "\<nu> x = (if x \<in> vars_rule (u\<^sub>2, v\<^sub>2) then \<tau>\<^sub>2 x else (sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>1) x)" for x

        have l\<^sub>1: "\<pi> \<bullet> u\<^sub>1 \<cdot> \<nu> = l\<^sub>1"
        proof -
          have "\<pi> \<bullet> u\<^sub>1 \<cdot> \<nu> = \<pi> \<bullet> u\<^sub>1 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>1"
            using disj
            by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def eqvt)
          then show ?thesis by simp
        qed
        have [simp]: "\<pi> \<bullet> v\<^sub>1 \<cdot> \<nu> = r\<^sub>1"
        proof -
          have "\<pi> \<bullet> v\<^sub>1 \<cdot> \<nu> = \<pi> \<bullet> v\<^sub>1 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>1"
            using disj
            by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def eqvt)
          then show ?thesis by simp
        qed
        have [simp]: "u\<^sub>2 \<cdot> \<nu> = l\<^sub>2"
        proof -
          have "u\<^sub>2 \<cdot> \<nu> = u\<^sub>2 \<cdot> \<tau>\<^sub>2" by (unfold term_subst_eq_conv) (auto simp: \<nu>_def vars_defs)
          then show ?thesis by simp
        qed
        then have [simp]: "u\<^sub>2 |_ p \<cdot> \<nu> = l\<^sub>2 |_ p"
          using True [THEN funposs_imp_poss] by (auto) (metis \<open>u\<^sub>2 \<cdot> \<nu> = l\<^sub>2\<close> subt_at_subst)

        have [simp]: "v\<^sub>2 \<cdot> \<nu> = r\<^sub>2"
        proof -
          have "v\<^sub>2 \<cdot> \<nu> = v\<^sub>2 \<cdot> \<tau>\<^sub>2" by (unfold term_subst_eq_conv) (auto simp: \<nu>_def vars_defs)
          then show ?thesis by simp
        qed

        have *: "\<pi> \<bullet> u\<^sub>1 \<cdot> \<nu> \<circ>\<^sub>s \<mu> = (u\<^sub>2 |_ p) \<cdot> \<nu> \<circ>\<^sub>s \<mu>" using \<open>l\<^sub>1 \<cdot> \<mu> = (l\<^sub>2 |_ p) \<cdot> \<mu>\<close> by (simp add: l\<^sub>1)
        then obtain \<mu>' where mgu: "mgu (\<pi> \<bullet> u\<^sub>1) (u\<^sub>2 |_ p) = Some \<mu>'"
          using mgu_complete by (auto simp: unifiers_def simp del: subst_subst_compose)
        from mgu_sound [OF this, THEN is_imgu_imp_is_mgu] and * obtain \<delta> where \<mu>': " \<nu> \<circ>\<^sub>s \<mu> = \<mu>' \<circ>\<^sub>s \<delta>"
          by (auto simp: is_mgu_def unifiers_def simp del: subst_subst_compose)

        have 1: "t \<cdot> \<mu>' \<cdot> \<delta> = t \<cdot> \<nu> \<cdot> \<mu>" for t
          by (fold subst_subst_compose) (simp add: \<mu>')
        let ?s = "(ctxt_of_pos_term p (u\<^sub>2 \<cdot> \<mu>'))\<langle>\<pi> \<bullet> v\<^sub>1 \<cdot> \<mu>'\<rangle>"
        let ?t = "v\<^sub>2 \<cdot> \<mu>'"
        have 2: "s \<cdot> \<sigma> = ?s \<cdot> \<delta> \<cdot> \<sigma>"
          using True [THEN funposs_imp_poss] and p [THEN funposs_imp_poss]
          by (simp add: s ctxt_of_pos_term_subst [symmetric] 1)
        have 3: "t \<cdot> \<sigma> = ?t \<cdot> \<delta> \<cdot> \<sigma>" by (simp add: t 1)

        have "\<not> \<pi> \<bullet> v\<^sub>1 \<cdot> \<mu>' \<succ> \<pi> \<bullet> u\<^sub>1 \<cdot> \<mu>'" (is "\<not> ?v \<cdot> _ \<succ> ?u \<cdot> _")
        proof
          presume "\<not> ?thesis"
          then have "?v \<cdot> \<nu> \<circ>\<^sub>s \<mu> \<succ> ?u \<cdot> \<nu> \<circ>\<^sub>s \<mu>" unfolding \<mu>' by (simp add: subst)
          then have "r\<^sub>1 \<cdot> \<mu> \<succ> l\<^sub>1 \<cdot> \<mu>" by (simp add: l\<^sub>1)
          moreover have "l\<^sub>1 \<succ> r\<^sub>1" using \<open>\<pi>\<^sub>1 \<bullet> l\<^sub>1 \<succ> \<pi>\<^sub>1 \<bullet> r\<^sub>1\<close> by (simp add: perm_less)
          ultimately show False using irrefl by (auto dest: subst trans)
        qed simp
        moreover have "\<not> v\<^sub>2 \<cdot> \<mu>' \<succ> u\<^sub>2 \<cdot> \<mu>'" (is "\<not> ?v \<cdot> _ \<succ> ?u \<cdot> _")
        proof
          presume "\<not> ?thesis"
          then have "?v \<cdot> \<nu> \<circ>\<^sub>s \<mu> \<succ> ?u \<cdot> \<nu> \<circ>\<^sub>s \<mu>" unfolding \<mu>' by (simp add: subst)
          then have "r\<^sub>2 \<cdot> \<mu> \<succ> l\<^sub>2 \<cdot> \<mu>" by simp
          moreover have "l\<^sub>2 \<succ> r\<^sub>2" using \<open>\<pi>\<^sub>2 \<bullet> l\<^sub>2 \<succ> \<pi>\<^sub>2 \<bullet> r\<^sub>2\<close> by (simp add: perm_less)
          ultimately show False using irrefl by (auto dest: subst trans)
        qed simp
        ultimately
        have "ooverlap {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>) (\<pi> \<bullet> u\<^sub>1, \<pi> \<bullet> v\<^sub>1) (u\<^sub>2, v\<^sub>2) p \<mu>' ?s ?t"
          using mgu and disj and rule1' and rule2'' and True
          by (auto simp: ooverlap_def perm_less)
        moreover
        have "\<forall>u \<lhd> \<pi> \<bullet> u\<^sub>1 \<cdot> \<mu>'. u \<in> NF ?RE"
          using NF and NF_ordstep_S_subset
          unfolding l\<^sub>1 [symmetric] and 1 [symmetric]
          by auto (meson NF_instance contra_subsetD instance_no_supt_imp_no_supt)
        ultimately have "(?s, ?t) \<in> PCP_ext (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)" by (force simp: PCP_ext_def)
        then show ?thesis
          unfolding 2 and 3 by blast
      qed }
    ultimately have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ?R" using o by (auto simp: overlap_def) }
  then show ?thesis
    using assms by (auto elim!: rstepE) blast+
qed

abbreviation "\<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c" where "\<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c \<equiv> \<E>\<^sup>\<leftrightarrow> \<inter> {\<succ>}"
definition \<R>':: "('a, 'b) term rel" where "\<R>' \<equiv> {(l, r) | l r. (l, r) \<in> dot (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c) \<and> (\<nexists>u l0 r0. (l, u) \<in> rstep {(l0, r0)} \<and> (l0, r0) \<in> \<S> \<and> l0 \<lhd>\<cdot> l)}"
definition \<E>' :: "('a, 'b) term rel" where "\<E>' \<equiv> {(some_NF (rstep \<R>') s, some_NF (rstep \<R>') t) | s t. (s, t) \<in> \<E> \<and> some_NF (rstep \<R>') s \<noteq> some_NF (rstep \<R>') t }"
definition "\<S>'":: "('a, 'b) term rel" where "\<S>' = \<R>' \<union> E_ord (op \<succ>) \<E>'"


lemma Esucc_E_ord: "\<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c \<subseteq> E_ord (op \<succ>) \<E>"
proof
  fix s t
  assume "(s, t) \<in> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c"
  note ord = this [unfolded Int_iff split]
  show "(s, t) \<in> E_ord (op \<succ>) \<E>" unfolding E_ord_def mem_Collect_eq
    by(rule exI [of _ s], rule exI [of _ t], rule exI [of _ Var], insert ord, auto)
qed

lemma encomp_perm: "u \<unlhd>\<cdot> s = (u \<unlhd>\<cdot> (\<pi> \<bullet> s))"
proof
  from term_apply_subst_Var_Rep_perm obtain \<sigma> where sub: "\<pi> \<bullet> s = s \<cdot> \<sigma>" by metis
  show "u \<unlhd>\<cdot> s \<Longrightarrow> u \<unlhd>\<cdot> \<pi> \<bullet> s" unfolding encompeq.simps sub
    by (metis subst_apply_term_ctxt_apply_distrib subst_subst)
next
  have "s = ( -\<pi>) \<bullet> \<pi> \<bullet> s" by auto
  with term_apply_subst_Var_Rep_perm obtain \<sigma> where sub: "s = (\<pi> \<bullet> s) \<cdot> \<sigma>" by metis
  show "u \<unlhd>\<cdot> \<pi> \<bullet> s \<Longrightarrow> u \<unlhd>\<cdot> s" unfolding encompeq.simps using sub
    by (metis subst_apply_term_ctxt_apply_distrib subst_subst)
qed

abbreviation "lexencomp \<equiv> lex_two {\<cdot>\<rhd>} {\<cdot>\<unrhd>}  {(a, b :: bool) . a \<and> \<not> b}"

lemma SN_lexencomp: "SN lexencomp"
proof-
  have t: "SN {(a, b). a \<and> \<not> b}" using SN_inv_image [OF SN_nat_gt, of "\<lambda>b. if b then 1 else 0"]
    by (metis (no_types, lifting) SN_onI mem_Collect_eq split_conv)
  then show ?thesis using SN_encomp encomp_order.less_le_trans encomp_order.le_less_trans
    by(intro lex_two, auto)
qed

lemma gterms_ER_conv_implies_S_conv:
  assumes RR_less: "RR \<subseteq> {\<succ>}"
  assumes ground: "ground s" "ground t"
  assumes conv: "(s, t) \<in> (rstep (RR \<union> EE\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" (is "_ \<in> ?R\<^sup>\<leftrightarrow>\<^sup>*")
  shows "(s, t) \<in> (GROUND (rstep (RR \<union> E_ord (op \<succ>) EE)))\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  have step_subset: "GROUND(rstep (RR \<union> EE\<^sup>\<leftrightarrow>)) \<subseteq> (GROUND (rstep (RR \<union> E_ord (op \<succ>) EE)))\<^sup>\<leftrightarrow>\<^sup>*"
  proof
    fix s t
    assume step: "(s, t) \<in> GROUND(rstep (RR \<union> EE\<^sup>\<leftrightarrow>))"
    then obtain C l r \<sigma> where C: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>" "(l, r) \<in> RR \<union> EE\<^sup>\<leftrightarrow>" "ground s" "ground t"
      unfolding GROUND_def by auto
    show "(s, t) \<in> (GROUND (rstep (RR \<union> E_ord (op \<succ>) EE)))\<^sup>\<leftrightarrow>\<^sup>*"
    proof (cases "(l, r) \<in> RR")
      case True
      with step have "(s, t) \<in> (rstep (RR \<union> E_ord (op \<succ>) EE))" unfolding C rstep_union by blast
      with C(4) C(5) show ?thesis unfolding GROUND_def by auto
    next
      case False
      with step rstep_union C(3) have in_E: "(l, r) \<in> EE\<^sup>\<leftrightarrow>" unfolding C by auto
      from C(4) C(5) have g': "ground (l \<cdot> \<sigma>)" "ground (r \<cdot> \<sigma>)" unfolding C by auto
      from gtotal [OF g'] consider " l \<cdot> \<sigma> = r \<cdot> \<sigma>" | "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma>" | "r \<cdot> \<sigma> \<succ> l \<cdot> \<sigma>" by auto
      then show ?thesis
      proof (cases)
        case 1
        then show ?thesis unfolding C by auto
      next
        case 2
        with in_E have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> E_ord (op \<succ>) EE" unfolding E_ord_def mem_Collect_eq by auto
        then have "(s, t) \<in> rstep (RR \<union> E_ord op \<succ> EE)" unfolding C by auto
        with C(4) C(5) show ?thesis unfolding GROUND_def by auto
      next
        case 3
        with in_E have "(r \<cdot> \<sigma>, l \<cdot> \<sigma>) \<in> E_ord (op \<succ>) EE" unfolding E_ord_def mem_Collect_eq by auto
        then have "(t, s) \<in> rstep (RR \<union> E_ord op \<succ> EE)" unfolding C by auto
        with C(4) C(5) show ?thesis unfolding GROUND_def by auto
      qed
    qed
  qed
  from gterm_conv_GROUND_conv assms have "(s, t) \<in> (GROUND (rstep (RR \<union> EE\<^sup>\<leftrightarrow>)))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  with conversion_mono [OF step_subset] show ?thesis unfolding conversion_conversion_idemp by auto
qed

lemma reduced_ground_complete:
  assumes "\<S> \<subseteq> {\<succ>}" and "GCR (rstep \<S>) \<and> SN (GROUND (rstep \<S>))"
  shows "GCR (rstep \<S>') \<and>
         SN (GROUND (rstep \<S>')) \<and>
         (GROUND (rstep \<S>))\<^sup>! = (GROUND (rstep \<S>'))\<^sup>!"
proof-
  let ?S = "\<R>' \<union> (E_ord (op \<succ>) \<E>')"
  have sub_dot: "\<R>' \<subseteq> dot (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)" unfolding \<R>'_def by fast
  with rstep_subset_less \<open>\<R> \<subseteq> {\<succ>}\<close> have **: "\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c \<subseteq> {\<succ>}" unfolding rstep_union by auto
  with rstep_subset_less \<open>\<R> \<subseteq> {\<succ>}\<close> have *: "rstep (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c) \<subseteq> {\<succ>}" unfolding rstep_union by auto
  with SN_less SN_subset have "SN (rstep (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c))" by auto
  interpret SN_trs "\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c" by standard fact

  { fix l r
    assume a: "(l, r) \<in> dot' (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)"
    then have "\<exists> r'. (l, r') \<in> (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c) \<and> r = some_NF (rstep (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)) r'"
      unfolding dot'_def mem_Collect_eq by fast
    with some_NF_rtrancl rstep.rule [OF a] have "(l, r) \<in> (rstep (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c))\<^sup>+"
      by (meson rstep.rule rtrancl_into_trancl2)
  }
  with rsteps_subset_less [OF **] rm_variants_subset have "dot (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c) \<subseteq> {\<succ>}"
    unfolding dot_def by blast
  with sub_dot rstep_subset_less have R'succ: "\<R>' \<subseteq> {\<succ>}" by auto
  with rstep_subset_less have "rstep \<R>' \<subseteq> {\<succ>}" unfolding E_ord_def by blast
  with SN_less SN_subset have "SN (rstep \<R>')" unfolding \<S>'_def by metis
  from R'succ rstep_subset_less have "rstep ?S \<subseteq> {\<succ>}" unfolding E_ord_def by blast
  with SN_less SN_subset GROUND_subset have x: "SN (GROUND (rstep \<S>'))" unfolding \<S>'_def by metis

  { fix s t
    assume "(s, t) \<in> GROUND (rstep \<S>)"
    then have g: "ground s" "ground t" "(s, t) \<in> rstep \<S>" unfolding GROUND_def by auto
    { fix l r
      assume in_S: "(l, r) \<in> \<S>"
      let ?orig = "\<lambda> rl. \<not> rl \<in> \<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c"
      from in_S have "\<exists> r'. (l, r') \<in> rstep \<S>'"
      proof (induct "(l, ?orig (l, r))" arbitrary: l r rule: SN_induct [OF SN_lexencomp])
        case (1 l)
        then show ?case
        proof (cases "(l, r) \<in> \<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c")
          case True
          note REsucc = this
          show ?thesis
          proof (cases "\<exists>u l0 r0. (l, u) \<in> rstep {(l0, r0)} \<and> (l0, r0) \<in> \<S> \<and> l0 \<lhd>\<cdot> l")
            case True
            then obtain u l0 r0 where step: "(l, u) \<in> rstep {(l0, r0)}" "(l0, r0) \<in> \<S>" and ll0: "l0 \<lhd>\<cdot> l" by blast
            then obtain C \<sigma> where C: "l = C\<langle>l0 \<cdot> \<sigma>\<rangle>" "(l0, r0) \<in> \<S>" using rstep.simps by auto
            from ll0 have "((l, ?orig (l, r)), (l0, ?orig (l0, r0))) \<in> lexencomp" by force
            with 1(1) [OF this] step have "\<exists> r0'. (l0, r0') \<in> rstep \<S>'" by auto
            with C(1) rstep.ctxt show ?thesis by force
          next
            case False
            from REsucc obtain u where in_dot: "(l, u) \<in> dot' (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)" unfolding dot'_def by auto
            from rm_variants_representative [OF this]
            obtain \<pi> l' r' where var: "(l', r') \<in> dot(\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)" "\<pi> \<bullet> (l, u) = (l', r')"
              unfolding dot_def by blast
            then have "\<pi> \<bullet> l = l' \<and> \<pi> \<bullet> u = r'" by (simp add: rule_pt.permute_prod.simps)
            then have lu: "l = ( - \<pi>) \<bullet> l'" "u = ( - \<pi>) \<bullet> r'" by auto
            { assume "\<exists>u l0 r0. (l', u) \<in> rstep {(l0, r0)} \<and> (l0, r0) \<in> \<S> \<and> l0 \<lhd>\<cdot> l'"
              then obtain v l0 r0 where step': "(l', v) \<in> rstep {(l0, r0)} "
                and in_S: "(l0, r0) \<in> \<S>" and enc: "l0 \<lhd>\<cdot> l'" by blast
              with rstep_r_p_s_imp_rstep have step: "(l', v) \<in> rstep {(l0, r0)}" by auto
              with permute_rstep lu have step: "(l, ( - \<pi>) \<bullet> v) \<in> rstep {(l0, r0)}" by auto
              from encomp_perm enc lu have enc: "l0 \<lhd>\<cdot> l" unfolding encomp_def
                by (metis encomp_def encomp_order.le_less_trans encompeq_refl)
              with False enc step in_S have False by blast
            }
            with var have "(l', r') \<in> \<R>'" unfolding \<R>'_def mem_Collect_eq by blast
            with var(2) perm_rstep_conv rstep.rule have "(l, u) \<in> rstep \<R>'" by metis
            then show ?thesis unfolding rstep_union \<S>'_def by fast
          qed
        next
          case False
          note unorientable = this
          with 1(2) obtain u v \<sigma> where uv: "(u, v) \<in> \<E>\<^sup>\<leftrightarrow>" "l = u \<cdot> \<sigma>" "r = v \<cdot> \<sigma>" "u \<cdot> \<sigma> \<succ> v \<cdot> \<sigma>"
            unfolding E_ord_def by blast
          let ?u = "some_NF (rstep \<R>') u"
          let ?v = "some_NF (rstep \<R>') v"
          interpret SN_R': SN_ars "rstep \<R>'" by standard fact
          from \<open>SN (rstep \<R>')\<close> have uu: "(u, ?u) \<in> (rstep \<R>')\<^sup>*"
            by (simp add: SN_ars.some_NF_rtrancl Normalization_Equivalence.SN_ars.intro)
          from \<open>SN (rstep \<R>')\<close> have vv: "(v, ?v) \<in> (rstep \<R>')\<^sup>*"
            by (simp add: SN_ars.some_NF_rtrancl Normalization_Equivalence.SN_ars.intro)
          show ?thesis
          proof (cases "?u = u")
            case True
            from vv rsteps_subset_less [OF \<open>\<R>' \<subseteq> {\<succ>}\<close>] have "v \<succeq> ?v"
              unfolding rtrancl_eq_or_trancl by auto
            with subst [of v ?v \<sigma>] trans [OF uv(4)] uv(4) have gt: "u \<cdot> \<sigma> \<succ> ?v \<cdot> \<sigma>" by auto
            with irrefl have "u \<noteq> ?v" by auto
            with uv(1) True have in_E': "(u, ?v) \<in> \<E>'\<^sup>\<leftrightarrow>"
              unfolding \<E>'_def Un_iff mem_Collect_eq converse_iff by metis
            from gt uv in_E' have "(l, ?v \<cdot> \<sigma>) \<in> E_ord op \<succ> \<E>'"
              unfolding E_ord_def mem_Collect_eq by blast
            with uv show ?thesis unfolding rstep_union \<S>'_def by blast
          next
            case False
            with uu have "(u, ?u) \<in> (rstep \<R>')\<^sup>+" by (simp add: rtrancl_eq_or_trancl)
            with tranclD obtain w where "(u, w) \<in> (rstep \<R>')" "(w, ?u) \<in> (rstep \<R>')\<^sup>*"
              by (meson tranclD)
            then obtain l0 r0 C \<tau> where l0r0: "u = C\<langle>l0 \<cdot> \<tau>\<rangle>" "w = C\<langle>r0 \<cdot> \<tau>\<rangle>" "(l0, r0) \<in> \<R>'" by auto
            then have "(l0, r0) \<in> dot (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)" unfolding \<R>'_def by auto
            with lhss_dot obtain r0' where oriented: "(l0, r0') \<in> \<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c" by fastforce
            from Esucc_E_ord oriented have in_S: "(l0, r0') \<in> \<S>" unfolding Un_iff by auto
            from uv(2) have "l \<cdot>\<unrhd> l0" unfolding encompeq.simps l0r0(1)
                subst_apply_term_ctxt_apply_distrib subst_subst by fast
            with unorientable oriented have "((l, ?orig (l, r)), (l0, ?orig (l0, r0'))) \<in> lexencomp" by simp
            from 1(1) [OF this in_S] show ?thesis unfolding uv l0r0 by fast
          qed
        qed
      qed
    }
    with g have "\<exists>t'. (s, t') \<in> rstep \<S>'" by fast
    then obtain t' where step: "(s, t') \<in> rstep \<S>'" by auto
    with \<open>rstep ?S \<subseteq> {\<succ>}\<close> ground_less [OF g(1)] have "ground t'" unfolding \<S>'_def by auto
    with g step have "\<exists>t'.(s, t') \<in> GROUND (rstep \<S>')" unfolding GROUND_def by auto
  }
  then have NFs: "NF (GROUND (rstep \<S>')) \<subseteq> NF (GROUND (rstep \<S>))" unfolding NF_def by auto

  have step_conv: "GROUND(rstep \<S>') \<subseteq> (GROUND (rstep \<S>))\<^sup>\<leftrightarrow>\<^sup>*"
  proof
    fix s t
    assume "(s, t) \<in> GROUND(rstep \<S>')"
    then have g: "ground s" "ground t" "(s, t) \<in> rstep \<S>'" unfolding GROUND_def by auto
    then obtain l r \<sigma> C where C: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>" "(l, r) \<in> \<S>'" by auto
    { fix l r
      assume "(l, r) \<in> \<R>'"
      with rm_variants_subset have "(l, r) \<in> dot' (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)" unfolding \<R>'_def mem_Collect_eq dot_def by fast
      then obtain r' where r: "r = some_NF (rstep (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)) r'" "(l, r') \<in> \<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c" unfolding dot'_def  by auto
      interpret REsucc: SN_ars "rstep (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)" by standard fact
      from \<open>SN (rstep (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c))\<close> have "(r', r) \<in> (rstep (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c))\<^sup>*" unfolding r
        by (simp add: SN_ars.some_NF_rtrancl Normalization_Equivalence.SN_ars.intro)
      with rstep_mono [of "\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c" "\<R> \<union> \<E>\<^sup>\<leftrightarrow>", THEN rtrancl_mono] have rr: "(r', r) \<in> (rstep  (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>*" by blast
      from r(2) have "(l, r') \<in> rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)" by fastforce
      with rr have "(l, r) \<in> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding conversion_def
        by (meson converse_rtrancl_into_rtrancl in_rtrancl_UnI)
    } note R'_RE = this
    {  fix s t
      assume "(s, t) \<in> rstep \<R>'"
      then obtain l r C \<sigma> where "(l, r) \<in> \<R>'" "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
      with R'_RE [OF this(1)] ctxt.closed_conversion [OF ctxt_closed_rstep, THEN ctxt.closedD]
        subst.closed_conversion [OF subst_closed_rstep, THEN subst.closedD] have "(s, t) \<in> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" by metis
    } note R'_step_RE = this
    then have "rstep \<R>' \<subseteq> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
    from conversion_conversion_idemp conversion_mono [OF this]
    have R'_steps_RE: "\<And> s t.(s, t) \<in> (rstep \<R>')\<^sup>* \<Longrightarrow> (s, t) \<in> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
    show "(s, t) \<in> (GROUND (rstep \<S>))\<^sup>\<leftrightarrow>\<^sup>*"
    proof (cases "(l, r) \<in> \<R>'")
      case True
      from R'_RE [OF this] ctxt.closed_conversion [OF ctxt_closed_rstep, THEN ctxt.closedD]
        subst.closed_conversion [OF subst_closed_rstep, THEN subst.closedD]
      have "(s, t) \<in> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding C by metis
      with gterms_ER_conv_implies_S_conv [OF \<open>\<R> \<subseteq> {\<succ>}\<close> g(1) g(2)] show ?thesis .
    next
      case False
      with C(3) [unfolded \<S>'_def] have "(l, r) \<in> E_ord op \<succ> \<E>'" by auto
      then obtain u v \<tau> where uv: "l = u\<cdot> \<tau>" "r = v \<cdot> \<tau>" "(u, v) \<in> \<E>'\<^sup>\<leftrightarrow>" "u\<cdot> \<tau> \<succ> v \<cdot> \<tau>"
        unfolding E_ord_def mem_Collect_eq by blast
      from uv(3) obtain u' v' where uv': "(u', v') \<in> \<E>\<^sup>\<leftrightarrow>" "u = some_NF (rstep \<R>') u'" "v = some_NF (rstep \<R>') v'"
        unfolding \<E>'_def Un_iff mem_Collect_eq by auto
      interpret SN_R': SN_ars "rstep \<R>'" by standard fact
      from \<open>SN (rstep \<R>')\<close> have "(u', u) \<in> (rstep \<R>')\<^sup>*" unfolding uv'
        by (simp add: SN_ars.some_NF_rtrancl Normalization_Equivalence.SN_ars.intro)
      with rsteps_closed_subst rsteps_closed_ctxt have "(C\<langle>u' \<cdot> \<tau> \<cdot> \<sigma>\<rangle>, C\<langle>u \<cdot> \<tau> \<cdot> \<sigma>\<rangle>) \<in> (rstep \<R>')\<^sup>*" by metis
      with R'_steps_RE conversion_inv have uconv: "(C\<langle>u \<cdot> \<tau> \<cdot> \<sigma>\<rangle>, C\<langle>u' \<cdot> \<tau> \<cdot> \<sigma>\<rangle>) \<in> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" by metis
      from \<open>SN (rstep \<R>')\<close> uv' have "(v', v) \<in> (rstep \<R>')\<^sup>*"
        by (simp add: SN_ars.some_NF_rtrancl Normalization_Equivalence.SN_ars.intro)
      with rsteps_closed_subst rsteps_closed_ctxt have "(C\<langle>v' \<cdot> \<tau> \<cdot> \<sigma>\<rangle>, C\<langle>v \<cdot> \<tau> \<cdot> \<sigma>\<rangle>) \<in> (rstep \<R>')\<^sup>*" by metis
      with R'_steps_RE have vconv: "(C\<langle>v' \<cdot> \<tau> \<cdot> \<sigma>\<rangle>, C\<langle>v \<cdot> \<tau> \<cdot> \<sigma>\<rangle>) \<in> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
      from uv' uv rstepI have "(C\<langle>u' \<cdot> \<tau> \<cdot>\<sigma>\<rangle>, C\<langle>v' \<cdot> \<tau> \<cdot>\<sigma>\<rangle>) \<in> rstep (\<E>\<^sup>\<leftrightarrow>)" by auto
      then have "(C\<langle>u' \<cdot> \<tau> \<cdot>\<sigma>\<rangle>, C\<langle>v' \<cdot> \<tau> \<cdot>\<sigma>\<rangle>) \<in> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding conversion_def using r_into_rtrancl
        by fastforce
      from conversion_trans' [OF uconv this, THEN conversion_trans'] vconv
      have "(s, t) \<in> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding C uv by auto
      with gterms_ER_conv_implies_S_conv [OF \<open>\<R> \<subseteq> {\<succ>}\<close> g(1) g(2)] show ?thesis .
    qed
  qed

  from assms have CR_SN: "CR (GROUND (rstep \<S>)) \<and> SN (GROUND (rstep \<S>))"
    by (meson GROUND_subset SN_subset)
  interpret complete_ars "GROUND (rstep \<S>)" by (standard, insert CR_SN, auto)
  from complete_NE_intro1 [OF step_conv x NFs] complete_ars.CR show ?thesis
    unfolding complete_ars_def Normalization_Equivalence.SN_ars_def by auto
qed
end

lemma SN_Sw_step: "SN (rstep S\<^sub>\<omega>)"
proof-
  have "S\<^sub>\<omega> \<subseteq> {\<succ>}" using Ri_less unfolding Sw_def R\<^sub>\<omega>_def by fast
  from SN_subset [OF SN_less] compatible_rstep_imp_less [OF this] show ?thesis by blast
qed

context
  assumes fair: "\<forall>(s, t)\<in> PCP_ext (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>). (s, t) \<in> (rstep E\<^sub>\<infinity>)\<^sup>\<leftrightarrow>"
begin

lemma msteps_imp_source_steps:
  assumes "\<forall> t \<in># M. s \<succ> t"
  shows "(mstep M \<R>)\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (\<Union>z \<in> {z. s \<succ> z}. slab \<R> z)\<^sup>\<leftrightarrow>\<^sup>*"
proof-
  have "mstep M \<R> \<subseteq> (\<Union>z \<in> {z. s \<succ> z}. slab \<R> z)"
  proof
    fix t u
    assume a: "(t, u) \<in> mstep M \<R>"
    then have tu: "(t, u) \<in> rstep \<R>" unfolding mstep_def mem_Collect_eq by auto
    from a have "\<exists> v \<in># M. v \<succeq> t" unfolding mstep_def mem_Collect_eq by auto
    with assms transD [OF trans_less] have st: "s \<succ> t" by blast
    from tu have "(t, u) \<in> slab \<R> t" unfolding source_step_def by auto
    with st show "(t, u) \<in> (\<Union>z \<in> {z. s \<succ> z}. slab \<R> z)" by auto
  qed
  from conversion_mono [OF this] show ?thesis by auto
qed

text \<open>Establish peak decreasingness of \<open>S\<^sub>\<omega>\<close> steps on ground terms for correctness result below.\<close>
lemma ground_Sw_peak_decreasing:
  assumes "ground s" and "(s, t) \<in> slab S\<^sub>\<omega> s" and "(s, u) \<in> slab S\<^sub>\<omega> s"
  shows "(t, u) \<in> (\<Union> v \<in> {v. ground v \<and> s \<succ> v}. slab S\<^sub>\<omega> v)\<^sup>\<leftrightarrow>\<^sup>*"
proof-
  from assms have rsteps: "(s, t) \<in> rstep S\<^sub>\<omega>" "(s, u) \<in> rstep S\<^sub>\<omega>" unfolding source_step_def by auto
  note vc = SN_imp_variable_condition [OF SN_Sw_step]
  from vc peak_imp_nabla2 [OF _ rsteps] have tu: "(t, u) \<in> nabla S\<^sub>\<omega> s ^^ 2" by auto
  have Sw_less: "S\<^sub>\<omega> \<subseteq> {\<succ>}" using Ri_less unfolding Sw_def R\<^sub>\<omega>_def by fast
  from compatible_rstep_imp_less [OF Sw_less] have Sw_step_less: "rstep S\<^sub>\<omega> \<subseteq> {\<succ>}" by blast
  let ?less_s_conv = "(\<Union>z\<in>{z. s \<succ> z}. slab S\<^sub>\<omega> z)\<^sup>\<leftrightarrow>\<^sup>*"
  let ?leq_conv = "\<lambda> v. (\<Union>z\<in>{z. v \<succeq> z}. slab S\<^sub>\<omega> z)\<^sup>\<leftrightarrow>\<^sup>*"
  { fix v w
    assume vw: "(v, w) \<in> nabla S\<^sub>\<omega> s" and "ground v" and "ground w" and "s \<succ> v" and "s \<succ> w"
    then have vw: "(v, w) \<in> (rstep S\<^sub>\<omega>)\<^sup>\<down> \<or> (v, w) \<in> (rstep (PCP S\<^sub>\<omega>))\<^sup>\<leftrightarrow>" unfolding nabla_def by auto
    { assume "(v, w) \<in> (rstep S\<^sub>\<omega>)\<^sup>\<down>"
      then obtain x where vx: "(v, x) \<in> (rstep S\<^sub>\<omega>)\<^sup>*" and wx: "(w, x) \<in> (rstep S\<^sub>\<omega>)\<^sup>*"
        unfolding join_def rtrancl_converse by blast
      from rsteps_slabI [OF vx _ Sw_step_less, of v] have "(v, x) \<in> ?leq_conv v" by auto
      from slab_conv_less_label [OF this] \<open>s \<succ> v\<close> have vx: "(v, x) \<in> ?less_s_conv" by auto
      from rsteps_slabI [OF wx _ Sw_step_less, of w] have "(w, x) \<in> ?leq_conv w" by auto
      from slab_conv_less_label [OF this] \<open>s \<succ> w\<close> have "(x, w) \<in> ?less_s_conv"
        unfolding conversion_inv by auto
      from conversion_trans' [OF vx this] have "(v, w) \<in> ?less_s_conv" by auto
    } note if_joinable = this
    { assume "(v, w) \<in> (rstep (PCP S\<^sub>\<omega>))\<^sup>\<leftrightarrow>"
      from PCP_xPCP [OF Rw_less \<open>ground v\<close> \<open>ground w\<close> this [unfolded Sw_def, folded E_ord_def]]
      have vw: "(v, w) \<in> (rstep S\<^sub>\<omega>)\<^sup>\<down> \<or> (v, w) \<in> (rstep (PCP_ext (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)))\<^sup>\<leftrightarrow>"
        unfolding Sw_def E_ord_def by auto
      { assume "(v, w) \<in> (rstep (PCP_ext (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)))\<^sup>\<leftrightarrow>"
        with fair rstep_rstep have "(v, w) \<in> (rstep E\<^sub>\<infinity>)\<^sup>\<leftrightarrow>" by fast
        from ground_Einf_Sw [OF this \<open>ground v\<close> \<open>ground w\<close>] have vw: "(v, w) \<in> (mstep {#v, w#} S\<^sub>\<omega>)\<^sup>\<leftrightarrow>\<^sup>*" .
        from \<open>s \<succ> v\<close> \<open>s \<succ> w\<close> have "\<forall> t \<in># {#v, w#}. s \<succ> t" by auto
        from vw msteps_imp_source_steps [OF this] have "(v, w) \<in> ?less_s_conv" by auto
      }
      with vw if_joinable have "(v, w) \<in> ?less_s_conv" by auto
    }
    with vw if_joinable have "(v, w) \<in> ?less_s_conv" by auto
  } note nabla_slab = this
  from tu obtain v where v: "(t, v) \<in> nabla S\<^sub>\<omega> s" "(v, u) \<in> nabla S\<^sub>\<omega> s" by auto
  note gimp = rstep_ground [OF _ \<open>ground s\<close>]
  from this [OF _ rsteps(1)] this [OF _ rsteps(2)] vc have "ground t" "ground u" by auto
  from v [unfolded nabla_def] rstep_ground [OF _ \<open>ground s\<close>] have sv: "(s, v) \<in> (rstep S\<^sub>\<omega>)\<^sup>+" by auto
  from rsteps_ground [OF _ \<open>ground s\<close> trancl_into_rtrancl [OF this]] vc have "ground v" by fast
  from sv rsteps_subset_less [OF Sw_less] have sv: "s \<succ> v" by auto
  from Sw_step_less rsteps have st: "s \<succ> t" and su: "s \<succ> u" by auto
  note tv = nabla_slab [OF v(1) \<open>ground t\<close> \<open>ground v\<close> st sv]
  note vu = nabla_slab [OF v(2) \<open>ground v\<close> \<open>ground u\<close> sv su]
  from ground_less [OF \<open>ground s\<close>] have "{v. ground v \<and> s \<succ> v} = {v. s \<succ> v}" by auto
  with conversion_trans' [OF tv vu] show ?thesis by auto
qed

lemma correctness_okb: "GCR (rstep S\<^sub>\<omega>) \<and> SN (rstep S\<^sub>\<omega>) \<and> (rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  interpret ars_peak_decreasing "slab S\<^sub>\<omega>" "{t. ground t}" "op \<succ>"
    by (unfold_locales, insert ground_Sw_peak_decreasing SN_less, auto)
  from CR have CR: "CR (\<Union>t \<in> {t. ground t}. slab S\<^sub>\<omega> t)" by auto
  from rstep_ground SN_imp_variable_condition [OF SN_Sw_step]
  have "\<And> s t. ground s \<Longrightarrow> (s, t) \<in> rstep S\<^sub>\<omega> \<Longrightarrow> ground t" by fastforce
  then have "(\<Union>t \<in> {t. ground t}. slab S\<^sub>\<omega> t) = GROUND (rstep S\<^sub>\<omega>)" unfolding source_step_def GROUND_def
    by blast
  with CR have "GCR (rstep S\<^sub>\<omega>)" unfolding GROUND_def source_step_def by auto
  with SN_Sw_step oKBi_conversion_ERw show ?thesis by auto
qed

end

end

end
