theory IA_Instance
  imports
  IA_Checker
  LTS_Termination_Prover
  LTS_Safety_Prover
begin

no_notation Simplex.curr_val_satisfies_state ("\<Turnstile>")

context IA begin

abbreviation ia_const where "ia_const x \<equiv> Fun (Const x) []" 
  
fun constant_exp where
  "constant_exp (Boolean True) = Fun Less [ia_const 0, ia_const 1]"
| "constant_exp (Boolean False) = Fun Less [ia_const 1, ia_const 0]"
| "constant_exp (Integer n) = ia_const n"

sublocale algebra_constant "interpret" "constant_exp"
proof
  show "vars_term (constant_exp d) = {}" for d by (cases d rule: constant_exp.cases, auto)
  show "\<lbrakk>constant_exp d\<rbrakk> \<alpha> = d" for d \<alpha> by (cases d rule: constant_exp.cases, auto)
qed

fun is_constant where
  "is_constant (Fun _ []) = True"
| "is_constant _ = False"

lemma is_constant:
  assumes "is_constant e" shows "\<lbrakk>e\<rbrakk>\<alpha> = \<lbrakk>e\<rbrakk>\<beta>"
proof(cases e)
  case (Fun f es) with assms show ?thesis by (cases es, auto)
qed (insert assms, auto)

sublocale art_checker \<Lambda> check_ia_clause check_ia_clause show_IA_exp show_IA_exp negate negate ..

sublocale pre_definability_checker \<Lambda> def_checker .

sublocale transition_definability_checker: transition_definability_checker \<Lambda> def_checker show_IA_exp show_IA_exp .. 

abbreviation "less_eq_formula s t \<equiv> Atom (Fun Lesseq [s,t])"
abbreviation "less_formula s t \<equiv> Atom (Fun Less [s,t])"

lemma less_eq_meaning:
  assumes \<alpha>: "assignment \<alpha>"
      and s: "s :\<^sub>f IntegerT"
      and t: "t :\<^sub>f IntegerT"
  shows "\<alpha> \<Turnstile> less_eq_formula s t \<longleftrightarrow> \<lbrakk>s\<rbrakk> \<alpha> \<le> \<lbrakk>t\<rbrakk> \<alpha>"
  using eval_types[OF s \<alpha>] eval_types[OF t \<alpha>] s t
  by (cases "\<lbrakk>s\<rbrakk>\<alpha>"; cases "\<lbrakk>t\<rbrakk>\<alpha>", auto simp: is_bool_def eval_true_def)

lemma less_meaning:
  assumes \<alpha>: "assignment \<alpha>"
      and s: "s :\<^sub>f IntegerT"
      and t: "t :\<^sub>f IntegerT"
  shows "\<alpha> \<Turnstile> less_formula s t \<longleftrightarrow> \<lbrakk>s\<rbrakk> \<alpha> < \<lbrakk>t\<rbrakk> \<alpha>"
  using eval_types[OF s \<alpha>] eval_types[OF t \<alpha>] s t
  by (cases "\<lbrakk>s\<rbrakk>\<alpha>"; cases "\<lbrakk>t\<rbrakk>\<alpha>", auto simp: is_bool_def eval_true_def)

(* lemma transition_removal is an instance of transition_removal_bounded *)
interpretation transition_removal: transition_removal_bounded \<Lambda> check_ia_clause show_IA_exp negate
  "op \<le> :: value \<Rightarrow> _ \<Rightarrow> _" "op <"
  IntegerT less_eq_formula less_formula is_constant
  by (unfold_locales; (fact less_eq_meaning less_meaning is_constant | simp add: All_less_Suc))


sublocale termination_checker: termination_checker \<Lambda> check_ia_clause check_ia_clause show_IA_exp show_IA_exp negate negate def_checker
  IntegerT "op \<le>" "op <" less_eq_formula less_formula is_constant
    by (unfold_locales)

definition check_termination where "check_termination \<equiv> termination_checker.check"

lemma check_termination: "isOK (check_termination Pi prf) \<Longrightarrow> lts_termination (lts_of Pi)"
  by (unfold check_termination_def, fact termination_checker.sound)

end

thm IA_logic.check_termination IA_logic.check_safety

declare IA_logic.constant_exp.simps[code]
  IA_logic.check_termination_def[code]
  IA_logic.check_safety_def[code]
  
end
