theory Integer_Arithmetic
imports
  Formula
  "$AFP/Show/Show_Instances"
  "$AFP/Deriving/Comparator_Generator/Compare_Generator"
  "$AFP/Certification_Monads/Check_Monad"
  Non_Inf_Orders
  "../Auxiliaries/Show_Polynomials"
begin
 
locale IA
begin
datatype ty = BoolT | IntegerT
datatype sig = Less | Lesseq | Sum nat | Const (the_Const: int) | Mult nat | Equals
datatype "value" = Integer (the_int: int) | Boolean (the_bool: bool)

end

type_synonym 'v IA_exp = "(IA.sig,'v,IA.ty) exp"
type_synonym 'v IA_formula = "(IA.sig,'v,IA.ty) exp formula"

interpretation IA_logic: IA.

derive "show" IA.sig
derive "show" IA.ty
derive compare_order IA.value

context IA begin

fun IA_type_of_fun :: "sig \<Rightarrow> ty list \<times> ty" where
  "IA_type_of_fun Less = ([IntegerT,IntegerT], BoolT)"
| "IA_type_of_fun Lesseq = ([IntegerT,IntegerT], BoolT)"
| "IA_type_of_fun Equals = ([IntegerT,IntegerT], BoolT)"
| "IA_type_of_fun (Sum n) = (replicate n IntegerT, IntegerT)"
| "IA_type_of_fun (Const _) = ([], IntegerT)"
| "IA_type_of_fun (Mult n) = (replicate n IntegerT, IntegerT)" 

fun IA_Values_of_type :: "ty \<Rightarrow> value set" where
  "IA_Values_of_type BoolT = range Boolean"
| "IA_Values_of_type IntegerT = range Integer"

fun "IA_interpret" :: "sig \<Rightarrow> value list \<Rightarrow> value" where
  "IA_interpret Less [x, y] = Boolean (the_int x < the_int y)"
| "IA_interpret Lesseq [x, y] = Boolean (the_int x \<le> the_int y)"
| "IA_interpret Equals [x, y] = Boolean (the_int x = the_int y)"
| "IA_interpret (Sum n) xs = Integer (sum_list (map the_int xs))"
| "IA_interpret (Const x) [] = Integer x"
| "IA_interpret (Mult n) xs = Integer (prod_list (map the_int xs))"
| "IA_interpret _ _ = undefined"

definition "\<Lambda> = \<lparr>
  type_of_fun = IA_type_of_fun,
  Values_of_type = IA_Values_of_type,
  Bool_types = {BoolT},
  to_bool = the_bool,
  interpret = IA_interpret
\<rparr>"

sublocale prelogic \<Lambda>.

lemma IA_fold: "IA_type_of_fun = type_of_fun"
  "IA_Values_of_type = Values_of_type"
  "the_bool = to_bool"
  "IA_interpret = interpret" by (auto simp: \<Lambda>_def)

lemma Bool_types[simp]: "Bool_types = {BoolT}" by (auto simp: \<Lambda>_def)

lemmas interpret_simps[simp] = IA_interpret.simps[unfolded IA_fold]
   and type_of_fun_simps[simp] = IA_type_of_fun.simps[unfolded IA_fold]
   and Values_of_type_simps[simp] = IA_Values_of_type.simps[unfolded IA_fold]
   and to_bool_simp[simp] = IA.value.sel(2)[unfolded IA_fold]

lemma mem_Values_of_typeE[elim]:
  assumes "v \<in> Values_of_type ty"
      and "\<And>i. ty = IntegerT \<Longrightarrow> v = Integer i \<Longrightarrow> thesis"
      and "\<And>b. ty = BoolT \<Longrightarrow> v = Boolean b \<Longrightarrow> thesis"
  shows "thesis" using assms by (cases ty, auto)

lemma mem_Values_of_typeD[dest]:
  assumes "v \<in> Values_of_type ty"
  shows "(ty = IntegerT \<longrightarrow> (\<exists>i. v = Integer i)) \<and> (ty = BoolT \<longrightarrow> (\<exists>b. v = Boolean b))"
  by (insert assms, auto)

lemma has_type_induct[consumes 1, case_names Less Lesseq Equals Sum Const Mult Var]:
  assumes "e :\<^sub>f ty"
  and LessCase: "\<And> ta tb.
       ta :\<^sub>f IntegerT \<Longrightarrow> tb :\<^sub>f IntegerT \<Longrightarrow> P ta IntegerT \<Longrightarrow> P tb IntegerT \<Longrightarrow> P (Fun Less [ta,tb]) BoolT"
  and LesseqCase: "\<And> ta tb.
       ta :\<^sub>f IntegerT \<Longrightarrow> tb :\<^sub>f IntegerT \<Longrightarrow> P ta IntegerT \<Longrightarrow> P tb IntegerT \<Longrightarrow> P (Fun Lesseq [ta,tb]) BoolT"
  and EqualsCase: "\<And> ta tb.
       ta :\<^sub>f IntegerT \<Longrightarrow> tb :\<^sub>f IntegerT \<Longrightarrow> P ta IntegerT \<Longrightarrow> P tb IntegerT \<Longrightarrow> P (Fun Equals [ta,tb]) BoolT"
  and SumCase: "\<And> ts n.
       (\<And> t. t \<in> set ts \<Longrightarrow> t :\<^sub>f IntegerT) \<Longrightarrow> (\<And> t. t \<in> set ts \<Longrightarrow> P t IntegerT) \<Longrightarrow> P (Fun (Sum n) ts) IntegerT"
  and ConstCase: "\<And> x. P (Fun (Const x) []) IntegerT"
  and MultCase: "\<And> ts n. 
       (\<And> t. t \<in> set ts \<Longrightarrow> t :\<^sub>f IntegerT) \<Longrightarrow> (\<And> t. t \<in> set ts \<Longrightarrow> P t IntegerT) \<Longrightarrow> P (Fun (Mult n) ts) IntegerT"
  and VarCase: "\<And> x ty. P (Var (x,ty)) ty"
  shows "P e ty"
using assms(1) proof(induct e ty rule: has_type.induct)
  case (1 x ty) with VarCase show ?case by(cases x, auto)
next
  case Fun: (2 f ls ty)
  show ?case
  proof(cases f)
    case [simp]: Less
    from Fun(2) obtain ta tb where [simp]: "ls = [ta,tb]" by (cases ls rule: list_3_cases, auto)
    from Fun(2) have at: "ta :\<^sub>f IntegerT" and bt: "tb :\<^sub>f IntegerT" and ty: "ty = BoolT" by auto
    with Fun.hyps[of 0] Fun.hyps[of 1] LessCase show ?thesis by auto
  next
    case [simp]: Lesseq
    from Fun(2) obtain ta tb where [simp]: "ls = [ta,tb]" by (cases ls rule: list_3_cases, auto)
    from Fun(2) have at: "ta :\<^sub>f IntegerT" and bt: "tb :\<^sub>f IntegerT" and ty: "ty = BoolT" by auto
    with Fun.hyps[of 0] Fun.hyps[of 1] LesseqCase show ?thesis by auto
  next
    case [simp]: Equals
    from Fun(2) obtain ta tb where [simp]: "ls = [ta,tb]" by (cases ls rule: list_3_cases, auto)
    from Fun(2) have at: "ta :\<^sub>f IntegerT" and bt: "tb :\<^sub>f IntegerT" and ty: "ty = BoolT" by auto
    with Fun.hyps[of 0] Fun.hyps[of 1] EqualsCase show ?thesis by auto
  next
    case [simp]: (Sum n)
    {
      fix l
      assume "l \<in> set ls" 
      then obtain i where l: "l = ls ! i" and i: "i < length ls" unfolding set_conv_nth by auto
      from Fun(2) l i have lty: "l :\<^sub>f IntegerT" by auto
      from Fun.hyps[OF i] lty i l Fun(2) have "P l IntegerT" by auto
      note lty this
    }
    from SumCase[OF this, of ls n] Fun(2) show ?thesis by auto
  next
    case [simp]: (Const x)
    with Fun have [simp]: "ls = []" by auto
    show ?thesis using Fun ConstCase by simp
  next
    case [simp]: (Mult n)
    {
      fix l
      assume "l \<in> set ls" 
      then obtain i where l: "l = ls ! i" and i: "i < length ls" unfolding set_conv_nth by auto
      from Fun(2) l i have lty: "l :\<^sub>f IntegerT" by auto
      from Fun.hyps[OF i] lty i l Fun(2) have "P l IntegerT" by auto
      note lty this
    }
    from MultCase[OF this, of ls n] Fun(2) show ?thesis by auto
  qed
qed

sublocale logic \<Lambda>
proof
  fix f ds
  show "length ds = length (param_types f) \<Longrightarrow>
       (\<And>i. i < length ds \<Longrightarrow> ds ! i \<in> Values_of_type (param_types f ! i)) \<Longrightarrow>
       interpret f ds \<in> Values_of_type (return_type f)"
  by (cases f; induct f ds rule: IA_interpret.induct, auto)
  show "Values_of_type ty \<noteq> {}" for ty by(cases ty, auto)
qed

lemmas [simp] = less_value_def compare_value_def lt_of_comp_def comparator_of_def

lemma less_IA_value_simps[simp]:
 "\<And>x y. Boolean x > Boolean y \<longleftrightarrow> x > y"
 "\<And>x y. Boolean x > Integer y"
 "\<And>x y. Integer x > Integer y \<longleftrightarrow> x > y"
 "\<And>x y. Integer x > Boolean y \<longleftrightarrow> False"
proof-
  show "Boolean x > Boolean y \<longleftrightarrow> x > y" for x y by (cases x; cases y, auto)
qed auto

lemmas [simp del] = less_value_def compare_value_def lt_of_comp_def comparator_of_def

lemma IA_value_less_eq[simp]: "(x :: value) \<ge> y \<longleftrightarrow> x = y \<or> x > y" by auto

end

instance "IA.value" :: non_inf_quasi_order
proof(intro_classes)
  fix f :: "nat \<Rightarrow> IA.value" and a
  assume "\<forall>i. f (Suc i) < f i"
  note chain = this[rule_format]
  have f0: "\<not> IA.Boolean True < f 0" by (cases "f 0", auto)
  from chain[of 0] have f1: "\<not> IA.Boolean False < f 1" by (cases "f 0"; cases "f 1", auto)
  have "\<exists>x. f (Suc (Suc n)) = IA.Integer x" for n
  proof(induct n)
    case 0
    from f1 chain[of 1] have "f (Suc (Suc 0)) < IA.Boolean False" by auto
    then show ?case by (cases "f (Suc (Suc 0))", auto)
  next
    case (Suc n)
    then obtain x where IH: "f (Suc (Suc n)) = IA.Integer x" by auto
    with chain[of "Suc (Suc n)"] show ?case by (cases "f (Suc (Suc (Suc n)))", auto)
  qed
  from choice[OF allI[OF this]] obtain g where g: "\<And>i. f (Suc (Suc i)) = IA.Integer (g i)" by force
  then have chaing: "g (Suc i) < g i" for i using chain[of "Suc (Suc i)"] by auto
  with non_inf_int_gt[unfolded non_inf_def]
  have "\<And>n. \<exists>i. \<not> n < g i" by blast
  with g have "\<exists>i. \<not> IA.Integer n < f (Suc (Suc i))" for n by auto
  then have *: "\<exists>i. \<not> IA.Integer n < f i" for n by auto
  show "\<exists>i. \<not> a < f i"
  proof(cases a)
    case a: (Boolean b) with f0 f1 show ?thesis by (cases b, auto)
  next
    case a: (Integer i) with * show ?thesis by auto
  qed
qed

subsection {* Non-term, polynomial version *}

type_synonym 'v ipoly = "('v,int)poly" 
  
datatype 'v poly_constraint = 
  Poly_Ge "'v ipoly" (* meaning: p \<ge> 0 *)
| Poly_Eq "'v ipoly" (* meaning: p = 0 *) 
 
context IA begin

fun shows_poly_constraint :: "'v :: show poly_constraint \<Rightarrow> shows" where 
  "shows_poly_constraint (Poly_Ge p) = (shows_poly p o shows '' >= 0'')"
| "shows_poly_constraint (Poly_Eq p) = (shows_poly p o shows '' = 0'')"

fun interpret_poly_constraint :: "('v \<Rightarrow> int) \<Rightarrow> 'v poly_constraint \<Rightarrow> bool" where
  "interpret_poly_constraint f (Poly_Ge p) = (eval_poly f p \<ge> 0)" 
| "interpret_poly_constraint f (Poly_Eq p) = (eval_poly f p = 0)" 
  
fun vars_poly_constraint :: "'v poly_constraint \<Rightarrow> 'v set" where
  "vars_poly_constraint (Poly_Ge p) = poly_vars p"
| "vars_poly_constraint (Poly_Eq p) = poly_vars p"

fun vars_poly_constraint_list :: "'v poly_constraint \<Rightarrow> 'v list" where
  "vars_poly_constraint_list (Poly_Ge p) = poly_vars_list p"
| "vars_poly_constraint_list (Poly_Eq p) = poly_vars_list p"
  
lemma vars_poly_constraint_list[simp]: "set (vars_poly_constraint_list c) = vars_poly_constraint c" 
  by (cases c, auto)
  
fun IA_exp_to_tpoly :: "'v IA_exp \<Rightarrow> ('v,int) tpoly" where
 "IA_exp_to_tpoly (Var (a,ty)) = PVar a" |
 "IA_exp_to_tpoly (Fun (Sum _) as) = PSum (map IA_exp_to_tpoly as)" |
 "IA_exp_to_tpoly (Fun (Const a) []) = PNum a" |
 "IA_exp_to_tpoly (Fun (Mult _) as) = PMult (map IA_exp_to_tpoly as)"

definition poly_const :: "'a :: zero \<Rightarrow> ('v,'a)poly" where
  "poly_const a = (if a = 0 then [] else [([],a)])" 
  
lemma poly_const[simp]: "eval_poly \<alpha> (poly_const a) = a" 
  unfolding poly_const_def by auto
 
definition poly_minus :: "('v,'a :: ring_1)poly \<Rightarrow> ('v,'a)poly \<Rightarrow> ('v,'a)poly" where
  "poly_minus f g = poly_add f (poly_mult (poly_const (-1)) g)" 

definition poly_one :: "('v,'a :: one)poly" where
  "poly_one = [([],1)]" 
  
  
lemma poly_minus[simp]: "eval_poly \<alpha> (poly_minus f g) = eval_poly \<alpha> f - eval_poly \<alpha> g" 
  unfolding poly_minus_def by simp

lemma poly_one[simp]: "eval_poly \<alpha> poly_one = 1" 
  unfolding poly_one_def by simp
    
definition IA_exp_to_poly :: "'v IA_exp \<Rightarrow> ('v,int) poly" where
  "IA_exp_to_poly = poly_of o IA_exp_to_tpoly" 


fun IA_exp_to_poly_constraint :: "'v IA_exp \<Rightarrow> 'v poly_constraint" where
  "IA_exp_to_poly_constraint (Fun Lesseq [a, b]) = Poly_Ge (poly_minus (IA_exp_to_poly b) (IA_exp_to_poly a))" 
| "IA_exp_to_poly_constraint (Fun Equals [a, b]) = Poly_Eq (poly_minus (IA_exp_to_poly b) (IA_exp_to_poly a))" 
| "IA_exp_to_poly_constraint (Fun Less [a, b]) = Poly_Ge (poly_minus (poly_minus (IA_exp_to_poly b) (IA_exp_to_poly a)) poly_one)" 

lemma IA_exp_to_tpoly:
  assumes \<alpha>: "assignment \<alpha>"
  and "e :\<^sub>f IntegerT"
  shows "\<lbrakk>e\<rbrakk>\<alpha> = Integer (eval_tpoly (\<lambda>a. the_int (\<alpha> (a, IntegerT))) (IA_exp_to_tpoly e))"
proof -
  have "e :\<^sub>f IntegerT \<Longrightarrow> ?thesis" 
  proof(induct e IntegerT rule:has_type_induct)
    case (Var x)
    from \<alpha> obtain i where "\<alpha> (x, IntegerT) = Integer i" by force
    then show ?case by auto
  next
    case (Sum es n)
    thus ?case by (auto, induct es, auto)
  next
    case (Mult es n)
    thus ?case by (auto, induct es, auto)
  qed auto
  thus ?thesis using assms by simp
qed
  
lemma IA_exp_to_poly:
  assumes \<alpha>: "assignment \<alpha>"
  and "e :\<^sub>f IntegerT"
shows "\<lbrakk>e\<rbrakk>\<alpha> = Integer (eval_poly (\<lambda>a. the_int (\<alpha> (a, IntegerT))) (IA_exp_to_poly e))"
  using IA_exp_to_tpoly[OF assms] unfolding IA_exp_to_poly_def o_def poly_of .
  
lemma IA_exp_to_poly_constraint: 
  assumes \<alpha>: "assignment \<alpha>"
  and e: "is_bool e"
  shows "to_bool (\<lbrakk>e\<rbrakk>\<alpha>) = (interpret_poly_constraint (\<lambda>a. the_int (\<alpha> (a, IntegerT))) (IA_exp_to_poly_constraint e))"
proof -
  have "e :\<^sub>f BoolT \<Longrightarrow> is_Fun e \<Longrightarrow> ?thesis"
    by (induct e BoolT rule: has_type_induct, insert IA_exp_to_poly[OF \<alpha>], auto)
  then show ?thesis using e[unfolded is_bool_def] by auto
qed
  
fun show_IA_exp where
  "show_IA_exp (Fun Less [s,t]) = show_IA_exp s o shows '' < '' o show_IA_exp t"
| "show_IA_exp (Fun Lesseq [s,t]) = show_IA_exp s o shows '' <= '' o show_IA_exp t"
| "show_IA_exp (Fun Equals [s,t]) = show_IA_exp s o shows '' = '' o show_IA_exp t"
| "show_IA_exp e = shows_poly (IA_exp_to_poly e)"

definition "show_IA_clause = shows_list_gen show_IA_exp ''FALSE'' [] '' || '' []" 

definition "show_IA_formula = shows_list_gen (shows_paren o show_IA_clause) ''TRUE'' [] '' && '' []" 
  
definition def_checker_main where
  "def_checker_main x ty l1 r1 l2 r2 = (is_bool (Fun Lesseq [l1, r1]) \<and>
    r1 :\<^sub>f ty \<and> is_bool (Fun Lesseq [l2, r2]) \<and> Var (x,ty) = l1 \<and> l1 = r2 \<and> l2 = r1 \<and> (x,ty) \<notin> vars_term r1)" 

definition def_checker_main' where
  "def_checker_main' x ty l1 r1 = (is_bool (Fun Equals [l1, r1]) \<and>
    r1 :\<^sub>f ty \<and> Var (x,ty) = l1 \<and> (x,ty) \<notin> vars_term r1)" 
  
definition def_checker_eq where "def_checker_eq x ty eq =
  (case eq of Atom e1 \<Rightarrow> (case e1 of Fun Equals [l1,r1] \<Rightarrow> 
              check (def_checker_main' x ty l1 r1 \<or> def_checker_main' x ty r1 l1)
              (shows ''Non-equation not supported'') | _ \<Rightarrow> error (shows ''Invalid literal''))
            | _ \<Rightarrow> error (shows ''Invalid clause''))" 
  
definition def_checker :: "(sig, 'v :: show, ty) definability_checker"
(* unfortunately, a nice pattern matching doesn't go well for proving... *)
where "def_checker x ty \<phi> \<equiv>
 (case \<phi> of Conjunction \<phi>s \<Rightarrow>
   (case \<phi>s of [] \<Rightarrow> succeed
    | [eq] \<Rightarrow> def_checker_eq x ty eq
    | [lt1, lt2] \<Rightarrow>
     (case lt1 of Atom e1 \<Rightarrow>
       (case e1 of Fun Lesseq [l1,r1] \<Rightarrow>
         (case lt2 of Atom e2 \<Rightarrow>
           (case e2 of Fun Lesseq [l2,r2] \<Rightarrow>
              check (def_checker_main x ty l1 r1 l2 r2 \<or> def_checker_main x ty l2 r2 l1 r1)
              (shows ''Non-equation not supported'')
             | _ \<Rightarrow> error (shows ''Invalid 2nd literal'')
           )
          | _ \<Rightarrow> error (shows ''Invalid 2nd clause'')
         )
        | _ \<Rightarrow> error (shows ''Invalid 1st literal'')
        )
      | _ \<Rightarrow> error (shows ''Invalid 1st clause'')
     )
   | _ \<Rightarrow> error (shows ''Conjunction of more than two'')
  )
 | _ \<Rightarrow> def_checker_eq x ty \<phi>)
 <+? (\<lambda> s. shows ''error in definibility checker for '' o shows x o shows '' on formula\<newline>  '' o shows_formula.shows_formula show_IA_exp \<phi> o shows_nl o s)"

sublocale definability_checker \<Lambda> "def_checker :: (_,'v :: show,_) definability_checker"
proof
  fix \<alpha> :: "'v \<times> ty \<Rightarrow> value" and x :: 'v and ty \<phi>
  assume ok: "isOK (def_checker x ty \<phi>)" and alpha: "assignment \<alpha>"
  note ok = ok[unfolded def_checker_def, simplified]
  { fix l1 r1 l2 r2
    assume "def_checker_main x ty l1 r1 l2 r2"
    note ok = this[unfolded def_checker_main_def]
    then have "is_bool (Fun Lesseq [l1, r1])" and "is_bool (Fun Lesseq [l2, r2])" by auto
    from ok have id: "l1 = Var (x,ty)" "r2 = Var (x,ty)" "l2 = r1" and fresh: "(x,ty) \<notin> vars_term r1" by auto
    from fresh have id2: "(\<lbrakk>r1\<rbrakk> (\<alpha>((x, ty) := a))) = (\<lbrakk>r1\<rbrakk> \<alpha>)" for a by simp
    from ok have "r1 :\<^sub>f ty" by auto
    with alpha have r1: "\<lbrakk>r1\<rbrakk> \<alpha> \<in> Values_of_type ty"
      using eval_types by auto
    have "(\<exists>a \<in> Values_of_type ty.
    \<alpha>((x, ty) := a) \<Turnstile> Atom (Fun Lesseq [l1, r1]) \<and>
    \<alpha>((x, ty) := a) \<Turnstile> Atom (Fun Lesseq [l2, r2]))"
      unfolding id using r1 by (intro bexI[of _ "\<lbrakk>r1\<rbrakk> \<alpha>"], auto simp: id2)
  } note * = this
  { fix l1 r1
    assume "def_checker_main' x ty l1 r1"
    note ok = this[unfolded def_checker_main'_def]
    then have "is_bool (Fun Equals [l1, r1])" by auto
    from ok have id: "l1 = Var (x,ty)" and fresh: "(x,ty) \<notin> vars_term r1" by auto
    from fresh have id2: "(\<lbrakk>r1\<rbrakk> (\<alpha>((x, ty) := a))) = (\<lbrakk>r1\<rbrakk> \<alpha>)" for a by simp
    from ok have "r1 :\<^sub>f ty" by auto
    with alpha have r1: "\<lbrakk>r1\<rbrakk> \<alpha> \<in> Values_of_type ty"
      using eval_types by auto
    have "\<exists>a \<in> Values_of_type ty.
      the_int (\<lbrakk>l1\<rbrakk> (\<alpha>((x, ty) := a))) = the_int (\<lbrakk>r1\<rbrakk> (\<alpha>((x, ty) := a)))"
      unfolding id using r1 by (intro bexI[of _ "\<lbrakk>r1\<rbrakk> \<alpha>"], auto simp: id2)
  } note ** = this
  {
    fix eq
    assume ok: "isOK(def_checker_eq x ty eq)"
    note ok = ok[unfolded def_checker_eq_def]
    from ok obtain l1 r1 where [simp]: "eq = Atom (Fun Equals [l1,r1])"
      by (auto split: list.splits sig.splits formula.splits term.splits)
    note **[dest!]
    note ok = ok[simplified]
    from ok have "\<exists>a\<in>Values_of_type ty. \<alpha>((x, ty) := a) \<Turnstile> eq" by force
  } note *** = this
  show "\<exists>a \<in> Values_of_type ty. \<alpha>((x, ty) := a) \<Turnstile> \<phi>"
  proof (cases \<phi>)
    case [simp]: (Conjunction \<phi>s)
    note ok = ok[simplified]
    show ?thesis
    proof (cases \<phi>s rule: list_4_cases)
      case Nil with ok show ?thesis by (cases ty, auto)
    next
      case [simp]: (2 cl1 cl2)
      note ok = ok[simplified]
      from ok obtain l1 r1 where [simp]: "cl1 = Atom (Fun Lesseq [l1,r1])"
        by (auto split: list.splits sig.splits formula.splits term.splits)
      with ok obtain l2 r2 where [simp]: "cl2 = Atom (Fun Lesseq [l2,r2])"
        by (auto split: list.splits sig.splits formula.splits term.splits)
      note ok = ok[simplified]
      note *[dest!]
      from ok show ?thesis by auto
    next
      case [simp]: (1 eq)
      show ?thesis using ok ***[of eq] by auto
    qed (insert ok, auto)
  next
    case [simp]: (Atom eq)
    with ok ***[of "Atom eq"] show ?thesis by auto
  qed (insert ok, auto simp: def_checker_eq_def)
qed
end

end