theory LTS_Parser
  imports Xmlt2 IA_Instance
begin

hide_const (open) name

(* move? *)

context
  fixes atom_parser :: "('f,'a,'t) exp formula xmlt2"
begin

partial_function (option) formula_parser :: "('f,'a,'t) exp formula xmlt2" where
  [code]: "formula_parser xml = (
  XMLbegin ''disjunction'' {
    fs \<leftarrow> xml_take_many 0 \<infinity> formula_parser; xml_return (Disjunction fs)
  } XMLor XMLbegin ''conjunction'' {
    fs \<leftarrow> xml_take_many 0 \<infinity> formula_parser; xml_return (Conjunction fs)
  } XMLor atom_parser) xml"

end

context
  fixes location_parser :: "'l xmlt2"
    and trans_parser :: "'tr xmlt2"
    and tatom_parser :: "('f,'v trans_var,'t) exp formula xmlt2"
begin

definition transition_parser :: "('tr \<times> ('f,'v,'t,'l) transition_rule) xmlt2" where
  "transition_parser \<equiv> XMLbegin ''transition'' {
    tr \<leftarrow> xml_take trans_parser;
    l \<leftarrow> xml_enter ''source'' (xml_take location_parser xml_return);
    r \<leftarrow> xml_enter ''target'' (xml_take location_parser xml_return);
    \<phi> \<leftarrow> xml_enter ''formula'' (xml_take (formula_parser tatom_parser) xml_return);
    xml_return (tr, Transition l r \<phi>)
  }" 

definition lts_parser :: "tag \<Rightarrow> ('f,'v,'t,'l,'tr) lts_impl xmlt2" where
  "lts_parser tag \<equiv> XMLbegin tag {
     i \<leftarrow> xml_enter ''initial'' (xml_take_many 1 \<infinity> location_parser xml_return);
     t \<leftarrow> xml_take_many 0 \<infinity> transition_parser;
    xml_return (Lts_Impl i t [])
   }"

definition "safety_input_parser \<equiv>
  XMLbegin ''ltsSafetyInput'' {
    lts \<leftarrow> xml_take (lts_parser ''lts'');
    err \<leftarrow> xml_enter ''error'' (xml_take_many 1 \<infinity> location_parser xml_return);
    xml_return (lts,err)
  }"

end

definition xml_change :: "'a xmlt2 \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> 'b xmlt2" where
  "xml_change p f x \<equiv> p x \<bind> (\<lambda>ret. case ret of Inl err \<Rightarrow> Some (Inl err) | Inr a \<Rightarrow> Some (Inr (f a)))"

context
  fixes art_node_id_parser :: "'n xmlst"
    and location_parser :: "'l xmlt2"
    and trans_parser :: "'tr xmlt2"
    and atom_parser :: "('f,'v,'t) exp formula xmlt2"
    and hint_parser :: "'h :: default xmlt2"
begin

definition formula_pos_parser :: "nat xmlt2" where [code]:
 "formula_pos_parser =
  XMLbegin ''conclusion'' {xml_return 0}
  XMLor XMLbegin ''assertion'' {xml_return 2}
  XMLor XMLbegin ''transition'' {xml_return 3}
  XMLor XMLbegin ''targetAssertion'' {xml_return 4}"

partial_function(option) hints_parser :: "'h hint xmlt2" where [code]:
 "hints_parser xml = (
  XMLbegin ''auto'' {
    xml_return default
  } XMLor XMLbegin ''distribute'' {
    pos \<leftarrow> xml_take formula_pos_parser;
    hints \<leftarrow> xml_take_many 0 \<infinity> hints_parser;
    xml_return (hint.Distribute pos hints)
  } XMLor XMLbegin ''erase'' {
    pos \<leftarrow> xml_take formula_pos_parser;
    hint \<leftarrow> xml_take hints_parser;
    xml_return (hint.Erase pos hint)
  } XMLor XMLbegin ''lexWeak'' {
    hints \<leftarrow> xml_take_many 0 \<infinity> hints_parser;
    xml_return (hint.LexWeak hints)
  } XMLor XMLbegin ''lexStrict'' {
    hints \<leftarrow> xml_take_many 0 \<infinity> hints_parser;
    xml_return (hint.LexStrict hints)
  } XMLor
    xml_change hint_parser hint.Base
) xml"

definition art_node_parser :: "(('f, 'v, 't, 'l, 'n, 'tr, 'h hint) art_node_impl \<times> 'n list) xmlt2" where
 "art_node_parser \<equiv> XMLbegin ''node'' {
    init \<leftarrow> xml_take_default False (xml_begin ''initial'' (xml_return True));
    nodeId \<leftarrow> xml_enter ''nodeId'' art_node_id_parser;
    invariant \<leftarrow> xml_enter ''invariant'' (xml_take (formula_parser atom_parser) xml_return);
    location \<leftarrow> xml_enter ''location'' (xml_take location_parser xml_return);
    edges \<leftarrow> xml_take (
      XMLbegin ''children'' {
        chs \<leftarrow> xml_take_many 0 \<infinity> (
          XMLbegin ''child'' {
            tr \<leftarrow> xml_take trans_parser;
            n \<leftarrow> xml_enter ''nodeId'' art_node_id_parser;
            h \<leftarrow> xml_take_default default (xml_begin ''hints'' (xml_take hints_parser xml_return));
            xml_return (tr, n, h)
          });
        xml_return (Children_Edge chs)
      } XMLor XMLbegin ''coverEdge'' {
          n \<leftarrow> xml_enter ''nodeId'' art_node_id_parser;
          h \<leftarrow> xml_take_default default (xml_begin ''hints'' (xml_take hints_parser xml_return));
          xml_return (Cover_Edge n h)
       });
    xml_return (Art_Node nodeId invariant location edges, if init then [nodeId] else [])
  }"

definition art_parser :: "(('f, 'v, 't, 'l, 'n, 'tr, 'h hint) art_impl) xmlt2" where
 "art_parser \<equiv> XMLbegin ''impact'' {
    init \<leftarrow> xml_take_default [] (xml_change (xml_begin ''initial'' art_node_id_parser) (\<lambda>x. [x]));
    pairs \<leftarrow> xml_enter ''nodes'' (xml_take_many 0 \<infinity> art_node_parser xml_return);
    let nodes = map fst pairs;
    let inits = init @ concat (map snd pairs);
    xml_return \<lparr> art_impl.initial_nodes = inits, nodes = nodes \<rparr>
  }"

end

context
  fixes location_parser :: "'l xmlt2"
    and trans_parser :: "'tr xmlt2"
    and atom_parser :: "('f,'v,'t) exp formula xmlt2"
    and hint_parser :: "'h :: default xmlt2"
begin

type_synonym art_node_id = string

abbreviation art where "art \<equiv> art_parser (xml_take_text xml_return) location_parser trans_parser atom_parser hint_parser"

definition "invariant_proof_parser \<equiv>
  xml_change art (\<lambda> prf i. Impact i prf)"

definition invariant_parser :: "('l \<times> ('f,'v,'t) exp formula) xmlt2" where
 "invariant_parser \<equiv> XMLbegin ''invariant'' {
    l \<leftarrow> xml_enter ''location'' (xml_take location_parser xml_return);
    \<phi> \<leftarrow> xml_enter ''formula'' (xml_take (formula_parser atom_parser) xml_return);
    xml_return (l,\<phi>)
  }"

definition invariants_parser :: "('l \<times> ('f,'v,'t) exp formula)list xmlt2" where
 "invariants_parser \<equiv> xml_begin ''invariants'' (xml_take_many 0 \<infinity> invariant_parser xml_return)"

definition safety_proof_parser :: "('f,'v,'t,'l,string,'tr,'h) safety_proof xmlt2" where
 "safety_proof_parser \<equiv> XMLbegin ''safetyViaInvariants'' {
    I \<leftarrow> xml_take invariants_parser;
    p \<leftarrow> xml_take invariant_proof_parser;
    xml_return (safety_proof.Invariant_Assertion (p I) safety_proof.Trivial)
  }"

end


fun cut_points_to_transitions :: "_ \<Rightarrow> _ \<Rightarrow> ('f,'v,'t,'l sharp,'tr sharp) transitions_impl" where
  "cut_points_to_transitions ts [] = ts"
| "cut_points_to_transitions ts ((l,tr,\<phi>)#cps) =
   cut_points_to_transitions ((Flat tr, Transition (Flat l) (Sharp l) \<phi>)#ts) cps"

definition xml_ignore :: "'a \<Rightarrow> 'a xmlt"
where
  "xml_ignore x xml \<equiv> return x"

context
  fixes location_id_parser :: "'l xmlst"
    and trans_id_parser :: "'tr xmlst"
    and exp_parser :: "('f,'v,'t) exp xmlt2"
    and atom_parser :: "('f,'v,'t) exp formula xmlt2"
    and tatom_parser :: "('f,'v trans_var,'t) exp formula xmlt2"
    and variable_parser :: "'v xmlt2"
    and hint_parser :: "'h :: default xmlt2"
    and type_parser :: "'t xmlt2"
    and dom_type :: "'t"
begin

definition "location_parser \<equiv> xml_begin ''locationId'' location_id_parser"

definition sharp_location_parser :: "'l sharp xmlt2" where
 "sharp_location_parser \<equiv>
    (xml_change (xml_begin ''locationDuplicate'' location_id_parser) Sharp)
    XMLor (xml_change (xml_begin ''locationId'' location_id_parser) Flat)"

abbreviation "tformula \<equiv> formula_parser tatom_parser"

context
  fixes trans_parser :: "'trr xmlt2"
begin

abbreviation "lts \<equiv> lts_parser sharp_location_parser trans_parser tatom_parser"

abbreviation "transition \<equiv> transition_parser sharp_location_parser trans_parser tatom_parser"

abbreviation "invariant_proof \<equiv> invariant_proof_parser sharp_location_parser trans_parser atom_parser hint_parser"

abbreviation "invariants \<equiv> invariants_parser sharp_location_parser atom_parser"

partial_function(option) cooperation_proof_parser where
 [code]: "cooperation_proof_parser xs = (
  xml_begin ''trivial'' (xml_return cooperation_proof.Trivial)
  XMLor XMLbegin ''newInvariants'' {
    i \<leftarrow> xml_take invariants;
    p \<leftarrow> xml_take invariant_proof;
    cp \<leftarrow> xml_take cooperation_proof_parser;
    xml_return (Invariants_Update (p i) cp)
  } XMLor XMLbegin ''transitionRemoval'' {
    rs \<leftarrow> XMLenter ''rankingFunctions'' {
      pairs \<leftarrow> xml_take_many 0 \<infinity> (XMLbegin ''rankingFunction'' {
        l \<leftarrow> xml_enter ''location'' (xml_take sharp_location_parser xml_return);
        es \<leftarrow> xml_enter ''expression'' (xml_take_many 1 \<infinity> exp_parser xml_return);
        xml_return (l,es)
      });
      xml_return pairs};
    bounds \<leftarrow> xml_enter ''bound'' (xml_take_many 0 \<infinity> exp_parser xml_return);
    removed \<leftarrow> xml_enter ''remove'' (xml_take_many 0 \<infinity> trans_parser xml_return);
    hinter \<leftarrow> xml_take_default (\<lambda>x. default) (XMLbegin ''hints'' {
      pairs \<leftarrow> xml_take_many 0 \<infinity> (XMLbegin ''hint'' {
        tr \<leftarrow> xml_take trans_parser;
        hint \<leftarrow> xml_take_default default (hints_parser hint_parser);
        xml_return (tr, hint)
      });
      xml_return (map_of_default default pairs)});
    inner \<leftarrow> xml_take cooperation_proof_parser;
    let rf = map_of_default [] rs;
    xml_return (Transition_Removal (Transition_removal_info rf removed dom_type bounds hinter) inner)
  } XMLor XMLbegin ''locationAddition'' {
    tr \<leftarrow> xml_take transition;
    prof \<leftarrow> xml_take cooperation_proof_parser;
    let (tr_id,tau) = tr;
    xml_return (Location_Addition (Location_Addition_Info (source tau) (target tau) tr_id tau) prof)
  } XMLor XMLbegin ''freshVariableAddition'' {
    x1 \<leftarrow> xml_take variable_parser;
    x2 \<leftarrow> xml_take type_parser;
    x3 \<leftarrow> XMLenter ''additionalFormulas'' {
      pairs \<leftarrow> xml_take_many 0 \<infinity> (XMLbegin ''additionalFormula'' {
         tr \<leftarrow> xml_take trans_parser;
         \<phi> \<leftarrow> xml_take tformula;
         xml_return (tr,\<phi>)
      });
      xml_return pairs
    };
    prof \<leftarrow> xml_take cooperation_proof_parser;
    xml_return (Fresh_Variable_Addition (Fresh_Variable_Addition_Info x1 x2 x3) prof)
  } XMLor XMLbegin ''cutTransitionSplit'' {
    ps \<leftarrow> xml_take_many 0 \<infinity> (XMLbegin ''cutTransitionsWithProof'' {
      cuts \<leftarrow> xml_enter ''cutTransitions'' (xml_take_many 0 \<infinity> trans_parser xml_return);
      prof \<leftarrow> xml_take cooperation_proof_parser;
      xml_return (cuts,prof)});
    xml_return (Cut_Transition_Split ps)
  } XMLor XMLbegin ''sccDecomposition'' {
    sccs \<leftarrow> xml_take_many 0 \<infinity> (XMLbegin ''sccWithProof'' {
      scc :: 'l sharp list \<leftarrow> xml_enter ''scc'' (xml_take_many 1 \<infinity> sharp_location_parser xml_return);
      prof \<leftarrow> xml_take cooperation_proof_parser;
      xml_return (scc,prof)});
    xml_return (Scc_Decomp sccs)
  }) xs"

end

definition "trans_id \<equiv> xml_begin ''transitionId'' trans_id_parser"


definition "sharp_trans_id =
  (xml_change (xml_begin ''transitionDuplicate'' trans_id_parser) Sharp)
  XMLor (xml_change (xml_begin ''transitionId'' trans_id_parser) Flat)"

definition "cutPoints_parser \<equiv> XMLbegin ''cutPoints'' {
  tuples \<leftarrow> xml_take_many 0 \<infinity> (XMLbegin ''cutPoint'' {
    l \<leftarrow> xml_enter ''locationId'' location_id_parser;
    tr \<leftarrow> xml_enter ''skipId'' trans_id_parser;
    \<phi> \<leftarrow> xml_enter ''skipFormula'' (xml_take tformula xml_return);
    xml_return (l,tr,\<phi>)
  });
  xml_return (cut_points_to_transitions [] tuples)
}"

partial_function (option) termination_proof_parser where [code]: "termination_proof_parser xml = (
  XMLbegin ''trivial'' {
    xml_return termination_proof.Trivial
  } XMLor XMLbegin ''newInvariants'' {
    i \<leftarrow> xml_take (invariants_parser location_parser atom_parser);
    p \<leftarrow> xml_take (invariant_proof_parser location_parser trans_id atom_parser hint_parser);
    cp \<leftarrow> xml_take termination_proof_parser;
    xml_return (Invariants_Update_LTS (p i) cp)
  } XMLor XMLbegin ''switchToCooperationTermination'' {
    cp \<leftarrow> xml_take cutPoints_parser;
    p \<leftarrow> xml_take (cooperation_proof_parser sharp_trans_id);
    xml_return (termination_proof.Via_Cooperation [(cp,p)])
  }
) xml"

end


type_synonym location_id = string
type_synonym transition_id = string
type_synonym variable_id = string


definition variable_parser :: "variable_id xmlt2" where
  "variable_parser = xml_begin ''variableId'' (xml_take_text xml_return)"

definition trans_var_parser :: "variable_id trans_var xmlt2" where
  "trans_var_parser =
    XMLbegin ''post'' {v \<leftarrow> xml_take variable_parser; xml_return (Post v)}
    XMLor XMLbegin ''aux'' {v \<leftarrow> xml_take_text; xml_return (Intermediate v)}
    XMLor (xml_change variable_parser Pre)"

context IA begin

definition "constant_parser = xml_begin ''constant'' (xml_take_int xml_return)"

partial_function(option) exp_parser' :: "'a xmlt2 \<Rightarrow> 'a IA_exp xmlt2" where [code]:
  "exp_parser' v xml = (
  XMLbegin ''product'' {
    exps \<leftarrow> xml_take_many 0 \<infinity> (exp_parser' v);
    xml_return (Fun (Mult (length exps)) exps)
  } XMLor XMLbegin ''sum'' {
    exps \<leftarrow> xml_take_many 0 \<infinity> (exp_parser' v);
    xml_return (Fun (Sum (length exps)) exps)
  } XMLor XMLbegin ''constant'' {
    i \<leftarrow> xml_take_int;
    xml_return (Fun (Const i) [])
  } XMLor xml_change v (\<lambda> v. Var (v, IntegerT))
  ) xml"

definition bexp_parser' :: "'a :: show xmlt2 \<Rightarrow> 'a IA_exp xmlt2" where
 "bexp_parser' v \<equiv>
  XMLbegin ''leq'' {
    l \<leftarrow> xml_take (exp_parser' v);
    r \<leftarrow> xml_take (exp_parser' v);
    xml_return (Fun Lesseq [l, r])
  } XMLor XMLbegin ''less'' {
    l \<leftarrow> xml_take (exp_parser' v);
    r \<leftarrow> xml_take (exp_parser' v);
    xml_return (Fun Less [l,r])
  } XMLor XMLbegin ''eq'' {
    l \<leftarrow> xml_take (exp_parser' v);
    r \<leftarrow> xml_take (exp_parser' v);
    xml_return (Fun Equals [l,r])
  }"

definition "exp_parser \<equiv> exp_parser' variable_parser" 
definition "atom_parser \<equiv> xml_change (bexp_parser' variable_parser) Atom"
definition "tatom_parser \<equiv> xml_change (bexp_parser' trans_var_parser) Atom"
definition "atom_parser' = xml_change (bexp_parser' variable_parser) Atom"
definition "tatom_parser' = xml_change (bexp_parser' trans_var_parser) Atom"

definition hint_parser :: "ia_hints xmlt2" where
  "hint_parser \<equiv> XMLbegin ''linearCombination'' {
      xs \<leftarrow> xml_take_many 0 \<infinity> (xml_begin ''constant'' (xml_take_int xml_return));
      xml_return (IA_Hints xs)
  }"

definition type_parser :: "ty xmlt2" where
  "type_parser = xml_begin ''int'' (xml_return IntegerT) XMLor xml_begin ''bool'' (xml_return BoolT)"


definition "lts_input_parser \<equiv>
  xmlt_of_xmlt2 (
    lts_parser
      (location_parser (xml_take_text xml_return))
      (trans_id (xml_take_text xml_return)) tatom_parser' ''lts'')"

definition "lts_safety_input_parser \<equiv> xmlt_of_xmlt2 (
  safety_input_parser
    (location_parser (xml_take_text xml_return))
    (trans_id (xml_take_text xml_return))
    tatom_parser'
)"

definition "lts_termination_proof_parser \<equiv> xmlt_of_xmlt2 (
  termination_proof_parser
    (xml_take_text xml_return)
    (xml_take_text xml_return)
    exp_parser atom_parser tatom_parser variable_parser  hint_parser type_parser IntegerT)"

definition "lts_safety_proof_parser \<equiv> xmlt_of_xmlt2 (
  safety_proof_parser
    (location_parser (xml_take_text xml_return))
    (trans_id (xml_take_text xml_return)) atom_parser hint_parser)"

end

end
